# Spec: https://github.com/KhronosGroup/glTF/blob/master/specification/2.0/README.md#asset
from __future__ import annotations

import base64
import io
import json
import logging
import struct
from collections import namedtuple
from pathlib import Path
from typing import Any, Optional, Union

import glm
import moderngl
import numpy
import numpy.typing as npt
from PIL import Image

import moderngl_window
from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.loaders.base import BaseLoader
from moderngl_window.loaders.texture import t2d
from moderngl_window.meta import SceneDescription, TextureDescription
from moderngl_window.opengl.vao import VAO
from moderngl_window.scene import Material, MaterialTexture, Mesh, Node, Scene

logger = logging.getLogger(__name__)

GLTF_MAGIC_HEADER = b"glTF"

# Texture wrap values
REPEAT = 10497
CLAMP_TO_EDGE = 33071
MIRRORED_REPEAT = 33648

# numpy dtype mapping
NP_COMPONENT_DTYPE = {
    5121: numpy.uint8,  # GL_UNSIGNED_BYTE
    5123: numpy.uint16,  # GL_UNSIGNED_SHORT
    5125: numpy.uint32,  # GL_UNSIGNED_INT
    5126: numpy.float32,  # GL_FLOAT
}

ComponentType = namedtuple("ComponentType", ["name", "value", "size"])

COMPONENT_TYPE = {
    5120: ComponentType("BYTE", 5120, 1),
    5121: ComponentType("UNSIGNED_BYTE", 5121, 1),
    5122: ComponentType("SHORT", 5122, 2),
    5123: ComponentType("UNSIGNED_SHORT", 5123, 2),
    5125: ComponentType("UNSIGNED_INT", 5125, 4),
    5126: ComponentType("FLOAT", 5126, 4),
}

# dtype to moderngl buffer format
DTYPE_BUFFER_TYPE = {
    numpy.uint8: "u1",  # GL_UNSIGNED_BYTE
    numpy.uint16: "u2",  # GL_UNSIGNED_SHORT
    numpy.uint32: "u4",  # GL_UNSIGNED_INT
    numpy.float32: "f4",  # GL_FLOAT
}

ACCESSOR_TYPE = {
    "SCALAR": 1,
    "VEC2": 2,
    "VEC3": 3,
    "VEC4": 4,
}


class Loader(BaseLoader):
    """Loader for GLTF 2.0 files"""

    kind = "gltf"
    file_extensions = [
        [".gltf"],
        [".glb"],
    ]
    #: Supported GLTF extensions
    #: https://github.com/KhronosGroup/glTF/tree/master/extensions
    supported_extensions: list[str] = [
        "KHR_draco_mesh_compression",
        "KHR_materials_unlit",
    ]

    meta: SceneDescription

    def __init__(self, meta: SceneDescription):
        """Initialize loading GLTF 2 scene.

        Supported formats:

        - gltf json format with external resources
        - gltf embedded buffers
        - glb Binary format
        """
        super().__init__(meta)
        self.scenes: list[Scene] = []
        self.nodes: list[Node] = []
        self.meshes: list[list[Mesh]] = []
        self.materials: list[Material] = []
        self.images: list[moderngl.Texture] = []
        self.samplers: list[moderngl.Sampler] = []
        self.textures: list[MaterialTexture] = []

        self.path: Optional[Path] = None
        self.scene: Scene
        self.gltf: GLTFMeta

    def load(self) -> Scene:
        """Load a GLTF 2 scene including referenced textures.

        Returns:
            Scene: The scene instance
        """
        assert self.meta.path is not None, "The path to this resource is empty"
        self.path = self.find_scene(self.meta.path)
        if not self.path:
            raise ImproperlyConfigured("Scene '{}' not found".format(self.meta.path))

        self.scene = Scene(str(self.path))

        # Load gltf json file
        if self.path.suffix == ".gltf":
            self.load_gltf()

        # Load binary gltf file
        if self.path.suffix == ".glb":
            self.load_glb()

        assert self.gltf is not None, "There is a problem with your file, could not load gltf"

        self.gltf.check_version()
        self.gltf.check_extensions(self.supported_extensions)
        self.load_images()
        self.load_samplers()
        self.load_textures()
        self.load_materials()
        self.load_meshes()
        self.load_nodes()

        self.scene.calc_scene_bbox()
        self.scene.prepare()

        return self.scene

    def load_gltf(self) -> None:
        """Loads a gltf json file parsing its contents"""
        with open(str(self.path)) as fd:
            self.gltf = GLTFMeta(str(self.path), json.load(fd), self.meta)

    def load_glb(self) -> None:
        """Loads a binary gltf file parsing its contents"""
        with open(str(self.path), "rb") as fd:
            # Check header
            magic = fd.read(4)
            if magic != GLTF_MAGIC_HEADER:
                raise ValueError(
                    "{} has incorrect header {!r} != {!r}".format(
                        self.path, magic, GLTF_MAGIC_HEADER
                    )
                )

            version = struct.unpack("<I", fd.read(4))[0]
            if version != 2:
                raise ValueError(f"{self.path} has unsupported version {version}")

            # Total file size including headers
            _ = struct.unpack("<I", fd.read(4))[0]  # noqa

            # Chunk 0 - json
            chunk_0_length = struct.unpack("<I", fd.read(4))[0]
            chunk_0_type = fd.read(4)
            if chunk_0_type != b"JSON":
                raise ValueError(
                    "Expected JSON chunk, not {!r} in file {}".format(chunk_0_type, self.path)
                )

            json_meta = fd.read(chunk_0_length).decode()

            # chunk 1 - binary buffer
            chunk_1_length = struct.unpack("<I", fd.read(4))[0]
            chunk_1_type = fd.read(4)
            if chunk_1_type != b"BIN\x00":
                raise ValueError(
                    "Expected BIN chunk, not {!r} in file {}".format(chunk_1_type, self.path)
                )

            self.gltf = GLTFMeta(
                str(self.path),
                json.loads(json_meta),
                self.meta,
                binary_buffer=fd.read(chunk_1_length),
            )

    def load_images(self) -> None:
        """Load images referenced in gltf metadata"""
        for image in self.gltf.images:
            self.images.append(image.load(self.path.parent))

    def load_samplers(self) -> None:
        """Load samplers referenced in gltf metadata"""
        for sampler in self.gltf.samplers:
            # Use a sane default sampler if the sampler data is empty
            # Samplers can simply just be json data: "{}"
            if sampler.minFilter is sampler.magFilter is None:
                self.samplers.append(
                    self.ctx.sampler(
                        filter=(moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR),
                        repeat_x=False,
                        repeat_y=False,
                        anisotropy=16.0,
                    )
                )
            else:
                self.samplers.append(
                    self.ctx.sampler(
                        filter=(sampler.minFilter, sampler.magFilter),
                        repeat_x=sampler.wrapS in [REPEAT, MIRRORED_REPEAT],
                        repeat_y=sampler.wrapT in [REPEAT, MIRRORED_REPEAT],
                        anisotropy=16.0,
                    )
                )

    def load_textures(self) -> None:
        """Load textures referenced in gltf metadata"""
        for texture_meta in self.gltf.textures:
            texture = MaterialTexture()

            if texture_meta.source is not None:
                texture.texture = self.images[texture_meta.source]

            if texture_meta.sampler is not None:
                texture.sampler = self.samplers[texture_meta.sampler]

            self.textures.append(texture)

    def load_meshes(self) -> None:
        """Load meshes referenced in gltf metadata"""
        for meta_mesh in self.gltf.meshes:
            # Returns a list of meshes
            meshes = meta_mesh.load(self.materials)
            self.meshes.append(meshes)

            for mesh in meshes:
                self.scene.meshes.append(mesh)

    def load_materials(self) -> None:
        """Load materials referenced in gltf metadata"""
        # Create material objects
        for meta_mat in self.gltf.materials:
            mat = Material(meta_mat.name)
            mat.color = meta_mat.baseColorFactor or (1.0, 1.0, 1.0, 1.0)
            mat.double_sided = meta_mat.doubleSided

            if meta_mat.baseColorTexture is not None:
                mat.mat_texture = self.textures[meta_mat.baseColorTexture["index"]]

            self.materials.append(mat)
            self.scene.materials.append(mat)

    def load_nodes(self) -> None:
        """Load nodes referenced in gltf metadata"""
        # Start with root nodes in the scene
        for node_id in self.gltf.scenes[0].nodes:
            node = self.load_node(self.gltf.nodes[node_id])
            self.scene.root_nodes.append(node)

    def load_node(self, meta: GLTFNode, parent: Optional[Node] = None) -> Node:
        """Load a single node"""
        # Create the node
        node = Node(name=meta.name, matrix=meta.matrix)
        self.scene.nodes.append(node)

        if meta.mesh is not None:
            # Since we split up meshes with multiple primitives, this can be a list
            # If only one mesh we set it on the node as normal
            if len(self.meshes[meta.mesh]) == 1:
                node.mesh = self.meshes[meta.mesh][0]
            # If multiple meshes we add them as new child node
            elif len(self.meshes[meta.mesh]) > 1:
                for mesh in self.meshes[meta.mesh]:
                    node.add_child(Node(mesh=mesh))

        if meta.camera is not None:
            # FIXME: Use a proper camera class
            node.camera = self.gltf.cameras[meta.camera]

        if parent:
            parent.add_child(node)

        # Follow children
        if meta.has_children:
            for node_id in meta.children:
                self.load_node(self.gltf.nodes[node_id], parent=node)

        return node


class GLTFMeta:
    """Container for gltf metadata"""

    def __init__(
        self,
        path: Union[Path, str],
        data: dict[Any, Any],
        meta: SceneDescription,
        binary_buffer: Optional[bytes] = None,
    ) -> None:
        """
        :param file: GLTF file name loaded
        :param data: Metadata (json loaded)
        :param binary_buffer: Binary buffer when loading glb files
        """
        self.path = Path(path) if isinstance(path, str) else path
        self.data = data
        self.meta = meta

        self.asset = GLTFAsset(data["asset"])
        self.materials = (
            [GLTFMaterial(m) for m in data["materials"]] if data.get("materials") else []
        )
        self.images = [GLTFImage(i) for i in data["images"]] if data.get("images") else []
        self.samplers = [GLTFSampler(s) for s in data["samplers"]] if data.get("samplers") else []
        self.textures = [GLTFTexture(t) for t in data["textures"]] if data.get("textures") else []
        self.scenes = [GLTFScene(s) for s in data["scenes"]] if data.get("scenes") else []
        self.nodes = [GLTFNode(n) for n in data["nodes"]] if data.get("nodes") else []
        self.meshes = [GLTFMesh(m, self.meta) for m in data["meshes"]] if data.get("meshes") else []
        self.cameras = [GLTFCamera(c) for c in data["cameras"]] if data.get("cameras") else []
        self.buffer_views = (
            [GLTFBufferView(i, v) for i, v in enumerate(data["bufferViews"])]
            if data.get("bufferViews")
            else []
        )
        self.buffers = (
            [GLTFBuffer(i, b, self.path.parent) for i, b in enumerate(data["buffers"])]
            if data.get("buffers")
            else []
        )
        self.accessors = (
            [GLTFAccessor(i, a) for i, a in enumerate(data["accessors"])]
            if data.get("accessors")
            else []
        )

        # glb files can contain buffer 0 data
        if binary_buffer:
            self.buffers[0].data = binary_buffer

        self._link_data()

        self.buffers_exist()
        self.images_exist()

    def _link_data(self) -> None:
        """Add references"""
        # accessors -> buffer_views -> buffers
        for acc in self.accessors:
            acc.bufferView = self.buffer_views[acc.bufferViewId]

        for buffer_view in self.buffer_views:
            buffer_view.buffer = self.buffers[buffer_view.bufferId]

        # Link accessors to mesh primitives
        for mesh in self.meshes:
            for primitive in mesh.primitives:
                # Link the bufferview for draco compressed buffers
                if primitive.extensions.get("KHR_draco_mesh_compression"):
                    ext = primitive.extensions["KHR_draco_mesh_compression"]
                    buffer_view_id = ext["bufferView"]
                    ext["bufferView"] = self.buffer_views[buffer_view_id]

                if primitive.indices is not None:
                    primitive.accessor = self.accessors[primitive.indices]
                for name, value in primitive.attributes.items():
                    primitive.attributes[name] = self.accessors[value]

        # Link buffer views to images
        for image in self.images:
            if image.bufferViewId is not None:
                image.bufferView = self.buffer_views[image.bufferViewId]

    @property
    def version(self) -> str:
        return self.asset.version

    def check_version(self, required: str = "2.0") -> None:
        if not self.version == required:
            msg = (
                f"GLTF Format version is not 2.0. Version states '{self.version}' "
                f"in file {self.path}"
            )
            raise ValueError(msg)

    def check_extensions(self, supported: list[str]) -> None:
        """
        "extensionsRequired": ["KHR_draco_mesh_compression"],
        "extensionsUsed": ["KHR_draco_mesh_compression"]
        """
        extReq = self.data.get("extensionsRequired")
        if extReq is not None:
            for ext in extReq:
                if ext not in supported:
                    raise ValueError(f"Extension '{ext}' not supported")
        extUse = self.data.get("extensionsUsed")
        if extUse is not None:
            for ext in extUse:
                if ext not in supported:
                    raise ValueError(f"Extension '{ext}' not supported")

    def buffers_exist(self) -> None:
        """Checks if the bin files referenced exist"""
        for buff in self.buffers:
            if not buff.is_separate_file:
                continue

            path = self.path.parent / buff.uri
            if not path.exists():
                raise FileNotFoundError(
                    f"Buffer {path} referenced in {self.path} not found"
                )

    def images_exist(self) -> None:
        """checks if the images references in textures exist"""
        pass


class GLTFAsset:
    """Asset Information"""

    def __init__(self, data: dict[str, Any]):
        self.version = data.get("version")
        self.generator = data.get("generator")
        self.copyright = data.get("copyright")


class GLTFMesh:
    class Primitives:
        def __init__(self, data: dict[str, Any]):
            self.attributes: dict[str, Any] = data.get("attributes")
            self.indices: int = data.get("indices")
            self.mode: int = data.get("mode")
            self.material = data.get("material")
            self.extensions = data.get("extensions", {})
            self.accessor: GLTFAccessor

    def __init__(self, data: dict[str, Any], meta: SceneDescription):
        self.meta = meta
        self.name = data.get("name", "")
        self.primitives = [GLTFMesh.Primitives(p) for p in data["primitives"]]

    def load(self, materials: list[Material]) -> list[Mesh]:
        name_map = {
            "POSITION": self.meta.attr_names.POSITION,
            "NORMAL": self.meta.attr_names.NORMAL,
            "TEXCOORD_0": self.meta.attr_names.TEXCOORD_0,
            "TANGENT": self.meta.attr_names.TANGENT,
            "JOINTS_0": self.meta.attr_names.JOINTS_0,
            "WEIGHTS_0": self.meta.attr_names.WEIGHTS_0,
            "COLOR_0": self.meta.attr_names.COLOR_0,
        }

        ctx = moderngl_window.ctx()
        meshes: list[Mesh] = []

        # FIXME: Split this up in methods
        # Read all primitives as separate meshes for now
        # According to the spec they can have different materials and vertex format
        for primitive in self.primitives:
            # Handle draco compressed meshes
            if primitive.extensions.get("KHR_draco_mesh_compression"):
                buffer_view = primitive.extensions["KHR_draco_mesh_compression"]["bufferView"]
                data = buffer_view.read_raw()
                try:
                    import DracoPy
                except ImportError:
                    raise ImportError("DracoPy is required to load draco compressed meshes")
                mesh = DracoPy.decode(data)

                attributes = {
                    'POSITION': {
                        'name': name_map["POSITION"],
                        'components': 3,
                        'type': 5126,
                    }
                }

                vao = VAO(self.name, mode=primitive.mode or moderngl.TRIANGLES)
                if mesh.faces is not None and mesh.faces.any():
                    vao.index_buffer(ctx.buffer(mesh.faces.astype("i4")))

                vao.buffer(mesh.points.astype("f4"), "3f", name_map["POSITION"])

                if mesh.tex_coord is not None and mesh.tex_coord.any():
                    vao.buffer(
                        ctx.buffer(mesh.tex_coord.astype("f4")),
                        "2f",
                        name_map["TEXCOORD_0"],
                    )
                    attributes["TEXCOORD_0"] = {
                        'name': name_map["TEXCOORD_0"],
                        'components': 2,
                        'type': 5126,
                    }
                if mesh.normals is not None and mesh.normals.any():
                    vao.buffer(ctx.buffer(mesh.normals.astype("f4")), "3f", name_map["NORMAL"])
                    attributes["NORMAL"] = {
                        'name': name_map["NORMAL"],
                        'components': 3,
                        'type': 5126,
                    }
                if mesh.colors is not None and mesh.colors.any():
                    vao.buffer(ctx.buffer(mesh.colors.astype("f4")), "4f", name_map["COLOR_0"])
                    attributes["COLOR_0"] = {
                        'name': name_map["COLOR_0"],
                        'components': 4,
                        'type': 5126,
                    }

                bbox_min, bbox_max = self.get_bbox(primitive)
                meshes.append(
                    Mesh(
                        self.name,
                        vao=vao,
                        attributes=attributes,
                        material=(
                            materials[primitive.material]
                            if primitive.material is not None else None
                        ),
                        bbox_min=bbox_min,
                        bbox_max=bbox_max,
                    )
                )
            else:
                vao = VAO(self.name, mode=primitive.mode or moderngl.TRIANGLES)

                # Index buffer
                component_type, index_vbo = self.load_indices(primitive)
                if index_vbo is not None:
                    vao.index_buffer(
                        ctx.buffer(index_vbo.tobytes()),
                        index_element_size=component_type.size,
                    )

                attributes = {}
                vbos = self.prepare_attrib_mapping(primitive)

                for vbo_info in vbos:
                    dtype, buffer = vbo_info.create()
                    vao.buffer(
                        buffer,
                        " ".join(
                            [
                                "{}{}".format(attr[1], DTYPE_BUFFER_TYPE[dtype])
                                for attr in vbo_info.attributes
                            ]
                        ),
                        [name_map[attr[0]] for attr in vbo_info.attributes],
                    )

                    for attr in vbo_info.attributes:
                        attributes[attr[0]] = {
                            "name": name_map[attr[0]],
                            "components": attr[1],
                            "type": vbo_info.component_type.value,
                        }

                bbox_min, bbox_max = self.get_bbox(primitive)
                meshes.append(
                    Mesh(
                        self.name,
                        vao=vao,
                        attributes=attributes,
                        material=(
                            materials[primitive.material]
                            if primitive.material is not None else None
                        ),
                        bbox_min=bbox_min,
                        bbox_max=bbox_max,
                    )
                )

        return meshes

    def load_indices(
        self, primitive: Primitives
    ) -> tuple[ComponentType, npt.NDArray[Any]] | tuple[None, None]:
        """Loads the index buffer / polygon list for a primitive"""
        if primitive.indices is None or primitive.accessor is None:
            return None, None

        _, component_type, buffer = primitive.accessor.read()
        return component_type, buffer

    def prepare_attrib_mapping(self, primitive: Primitives) -> list[VBOInfo]:
        """Pre-parse buffer mappings for each VBO to detect interleaved data for a primitive"""
        buffer_info: list[VBOInfo] = []
        for name, accessor in primitive.attributes.items():
            info = VBOInfo(*accessor.info())
            info.attributes.append((name, info.components))

            if buffer_info and buffer_info[-1].buffer_view == info.buffer_view:
                if buffer_info[-1].interleaves(info):
                    buffer_info[-1].merge(info)
                    continue

            buffer_info.append(info)

        return buffer_info

    def get_bbox(self, primitive: Primitives) -> tuple[glm.vec3, glm.vec3]:
        """Get the bounding box for the mesh"""
        accessor = primitive.attributes.get("POSITION")
        return glm.vec3(accessor.min), glm.vec3(accessor.max)


class VBOInfo:
    """Resolved data about each VBO"""

    def __init__(
        self,
        buffer: Optional[GLTFBuffer] = None,
        buffer_view: Optional[GLTFBuffer] = None,
        byte_length: int = 0,
        byte_offset: int = 0,
        component_type: ComponentType = ComponentType("", 0, 0),
        components: int = 0,
        count: int = 0,
    ):
        self.buffer = buffer  # reference to the buffer
        self.buffer_view = buffer_view
        self.byte_length = byte_length  # Raw byte buffer length
        self.byte_offset = byte_offset  # Raw byte offset
        self.component_type = component_type  # Datatype of each component
        self.components = components
        self.count = count  # number of elements of the component type size

        # list of (name, components) tuples
        self.attributes: list[Any] = []

    def interleaves(self, info: VBOInfo) -> bool:
        """Does the buffer interleave with this one?"""
        return bool(info.byte_offset == (self.component_type.size * self.components))

    def merge(self, info: VBOInfo) -> None:
        # NOTE: byte length is the same
        self.components += info.components
        self.attributes += info.attributes

    def create(self) -> tuple[type[object], npt.NDArray[Any]]:
        """Create the VBO"""
        assert self.buffer is not None, "No buffer defined"
        dtype = NP_COMPONENT_DTYPE[self.component_type.value]
        data = numpy.frombuffer(
            self.buffer.read(byte_length=self.byte_length, byte_offset=self.byte_offset),
            count=self.count * self.components,
            dtype=dtype,
        )
        return dtype, data

    def __str__(self) -> str:
        assert self.buffer is not None, "No buffer defined"
        assert self.buffer_view is not None, "No buffer_view defined"
        return (
            "VBOInfo<buffer={}, buffer_view={},\n"
            "        length={}, offset={}, count={}\n"
            "        component_type={}, components={}, \n"
            "        attribs={}".format(
                self.buffer.id,
                self.buffer_view.id,
                self.byte_length,
                self.byte_offset,
                self.count,
                self.component_type.value,
                self.components,
                self.attributes,
            )
        )

    def __repr__(self) -> str:
        return str(self)


class GLTFAccessor:
    def __init__(self, accessor_id: int, data: dict[str, Any]):
        self.id = accessor_id
        self.bufferViewId = data.get("bufferView", 0)
        self.bufferView: GLTFBufferView
        self.byteOffset = data.get("byteOffset", 0)
        self.componentType = COMPONENT_TYPE[data["componentType"]]
        self.count = data.get("count", 1)
        self.type: str = data.get("type", "")

        self.min = numpy.array(data.get("min") or [-0.5, -0.5, -0.5], dtype="f4")
        self.max = numpy.array(data.get("max") or [0.5, 0.5, 0.5], dtype="f4")

    def read(self) -> tuple[int, ComponentType, npt.NDArray[Any]]:
        """
        Reads buffer data

        Return:
            component count, component type, data
        """
        # ComponentType helps us determine the datatype
        dtype = NP_COMPONENT_DTYPE[self.componentType.value]
        return (
            ACCESSOR_TYPE[self.type],
            self.componentType,
            self.bufferView.read(
                byte_offset=self.byteOffset,
                dtype=dtype,
                count=self.count * ACCESSOR_TYPE[self.type],
            ),
        )

    def read_raw(self) -> bytes:
        """
        Read the raw bytes. Useful for draco compressed meshes or any data that
        is not a simple vertex buffer.
        """
        return self.bufferView.read_raw(byte_offset=self.byteOffset)

    def info(self) -> tuple[GLTFBuffer, GLTFBufferView, int, int, ComponentType, int, int]:
        """
        Get underlying buffer info for this accessor

        Return:
            buffer, byte_length, byte_offset, component_type, count
        """
        buffer, byte_length, byte_offset = self.bufferView.info(byte_offset=self.byteOffset)
        return (
            buffer,
            self.bufferView,
            byte_length,
            byte_offset,
            self.componentType,
            ACCESSOR_TYPE[self.type],
            self.count,
        )

    def __str__(self) -> str:
        return "Accessor<id={}, bufferView={}, byteOffset={}, componentType={}, count={}>".format(
            self.id, self.bufferViewId, self.byteOffset, self.componentType.name, self.count
        )


class GLTFBufferView:
    def __init__(self, view_id: int, data: dict[str, Any]):
        self.id = view_id
        self.bufferId = data.get("buffer", 0)
        self.buffer: GLTFBuffer
        self.byteOffset = data.get("byteOffset", 0)
        self.byteLength = data.get("byteLength", 0)
        self.byteStride = data.get("byteStride", 0)
        # Valid: 34962 (ARRAY_BUFFER) and 34963 (ELEMENT_ARRAY_BUFFER) or None

    def read(
        self, byte_offset: int = 0, dtype: Optional[type[object]] = None, count: int = 0
    ) -> npt.NDArray[Any]:
        data = self.buffer.read(
            byte_offset=byte_offset + self.byteOffset,
            byte_length=self.byteLength,
        )
        vbo = numpy.frombuffer(data, count=count, dtype=dtype)
        return vbo

    def read_raw(self, byte_offset: int = 0) -> bytes:
        return self.buffer.read(
            byte_offset=self.byteOffset + byte_offset,
            byte_length=self.byteLength,
        )

    def info(self, byte_offset: int = 0) -> tuple[GLTFBuffer, int, int]:
        """
        Get the underlying buffer info
        :param byte_offset: byte offset from accessor
        :return: buffer, byte_length, byte_offset
        """
        return self.buffer, self.byteLength, byte_offset + self.byteOffset

    def __str__(self) -> str:
        return "BufferView<id={}, buffer={}, byteOffset={}, byteLength={}>".format(
            self.id, self.bufferId, self.byteOffset, self.byteLength
        )


class GLTFBuffer:
    def __init__(self, buffer_id: int, data: dict[str, str], path: Path):
        self.id = buffer_id
        self.path = path
        self.byteLength = data.get("byteLength")
        uri = data.get("uri")
        if uri is None:
            uri = ""
        self.uri = uri
        self.data = b""

    @property
    def has_data_uri(self) -> bool:
        """Is data embedded in json?"""
        if self.uri == "":
            return False

        return self.uri.startswith("data:")

    @property
    def is_separate_file(self) -> bool:
        """Buffer represents an independent bin file?"""
        return self.uri is not None and not self.has_data_uri

    def open(self) -> None:
        if self.data != b"":
            return

        if self.has_data_uri:
            self.data = base64.b64decode(self.uri[self.uri.find(",") + 1 :])
            return

        with open(str(self.path / (self.uri if self.uri is not None else "")), "rb") as fd:
            self.data = fd.read()

    def read(self, byte_offset: int = 0, byte_length: int = 0) -> bytes:
        self.open()
        return self.data[byte_offset : byte_offset + byte_length]


class GLTFScene:
    def __init__(self, data: dict[str, list[int]]):
        self.nodes = data["nodes"]


class GLTFNode:
    def __init__(self, data: dict[str, Any]) -> None:
        self.name = data.get("name")
        self.children = data.get("children")
        self.mesh = data.get("mesh")
        self.camera = data.get("camera")

        _matrix = data.get("matrix")
        self.matrix = glm.mat4(*_matrix) if _matrix is not None else glm.mat4()

        self.translation = data.get("translation")
        self.rotation = data.get("rotation")
        self.scale = data.get("scale")

        if self.translation is not None:
            self.matrix = glm.translate(self.matrix, glm.vec3(*self.translation))

        if self.rotation is not None:
            quat = glm.quat(
                x=self.rotation[0],
                y=self.rotation[1],
                z=self.rotation[2],
                w=self.rotation[3],
            )
            self.matrix = self.matrix * glm.mat4_cast(quat)

        if self.scale is not None:
            self.matrix = glm.scale(self.matrix, glm.vec3(*self.scale))

    @property
    def has_children(self) -> bool:
        return self.children is not None and len(self.children) > 0

    @property
    def is_resource_node(self) -> bool:
        """Is this just a reference node to a resource?"""
        return self.camera is not None or self.mesh is not None


class GLTFMaterial:
    def __init__(self, data: dict[str, Any]):
        self.name = data.get("name")
        # Defaults to true if not defined
        self.doubleSided = data.get("doubleSided") or True

        pbr = data["pbrMetallicRoughness"]
        self.baseColorFactor = pbr.get("baseColorFactor")
        self.baseColorTexture = pbr.get("baseColorTexture")
        self.metallicFactor = pbr.get("metallicFactor")
        self.emissiveFactor = data.get("emissiveFactor")


class GLTFImage:
    """
    Represent texture data.
    May be a file, embedded data or pointer to data in bufferview
    """

    def __init__(self, data: dict[str, Any]):
        self.uri = data.get("uri")
        self.bufferViewId = data.get("bufferView")
        self.bufferView = None
        self.mimeType = data.get("mimeType")

    def load(self, path: Path) -> moderngl.Texture:
        # data:image/png;base64,iVBOR

        # Image is stored in bufferView
        if self.bufferView is not None:
            image = Image.open(io.BytesIO(self.bufferView.read_raw()))
        # Image is embedded
        elif self.uri and self.uri.startswith("data:"):
            data = self.uri[self.uri.find(",") + 1 :]
            image = Image.open(io.BytesIO(base64.b64decode(data)))
            logger.info("Loading embedded image")
        else:
            path = path / Path(self.uri if self.uri is not None else "")
            logger.info("Loading: %s", self.uri)
            image = Image.open(path)

        texture = t2d.Loader(
            TextureDescription(
                label="gltf",
                image=image,
                flip=False,
                mipmap=True,
                anisotropy=16.0,
            )
        ).load()

        return texture


class GLTFTexture:
    def __init__(self, data: dict[str, int]):
        self.sampler: Optional[int] = data.get("sampler")
        self.source: Optional[int] = data.get("source")


class GLTFSampler:
    def __init__(self, data):
        self.magFilter = data.get("magFilter")
        self.minFilter = data.get("minFilter")
        self.wrapS = data.get("wrapS")
        self.wrapT = data.get("wrapT")


class GLTFCamera:
    def __init__(self, data: dict[str, str]):
        self.data = data
        # "perspective": {
        #     "aspectRatio": 1.0,
        #     "yfov": 0.266482561826706,
        #     "zfar": 1000000.0,
        #     "znear": 0.04999999701976776
        # },
        # "type": "perspective"



# flake8: noqa E741
import pygame

from moderngl_window.context.base import BaseKeys


class Keys(BaseKeys):
    """
    Namespace mapping pygame2 specific key constants
    """

    ACTION_PRESS = pygame.KEYDOWN
    ACTION_RELEASE = pygame.KEYUP

    ESCAPE = pygame.K_ESCAPE
    SPACE = pygame.K_SPACE
    ENTER = pygame.K_RETURN
    PAGE_UP = pygame.K_PAGEUP
    PAGE_DOWN = pygame.K_PAGEDOWN
    LEFT = pygame.K_LEFT
    RIGHT = pygame.K_RIGHT
    UP = pygame.K_UP
    DOWN = pygame.K_DOWN
    LEFT_SHIFT = pygame.K_LSHIFT
    RIGHT_SHIFT = pygame.K_RSHIFT
    LEFT_CTRL = pygame.K_LCTRL

    TAB = pygame.K_TAB
    COMMA = pygame.K_COMMA
    MINUS = pygame.K_MINUS
    PERIOD = pygame.K_PERIOD
    SLASH = pygame.K_SLASH
    SEMICOLON = pygame.K_SEMICOLON
    EQUAL = pygame.K_EQUALS
    LEFT_BRACKET = pygame.K_LEFTBRACKET
    RIGHT_BRACKET = pygame.K_RIGHTBRACKET
    BACKSLASH = pygame.K_BACKSLASH
    BACKSPACE = pygame.K_BACKSPACE
    INSERT = pygame.K_INSERT
    DELETE = pygame.K_DELETE
    HOME = pygame.K_HOME
    END = pygame.K_END
    CAPS_LOCK = pygame.K_CAPSLOCK

    F1 = pygame.K_F1
    F2 = pygame.K_F2
    F3 = pygame.K_F3
    F4 = pygame.K_F4
    F5 = pygame.K_F5
    F6 = pygame.K_F6
    F7 = pygame.K_F7
    F8 = pygame.K_F8
    F9 = pygame.K_F9
    F10 = pygame.K_F10
    F11 = pygame.K_F11
    F12 = pygame.K_F12

    NUMBER_0 = pygame.K_0
    NUMBER_1 = pygame.K_1
    NUMBER_2 = pygame.K_2
    NUMBER_3 = pygame.K_3
    NUMBER_4 = pygame.K_4
    NUMBER_5 = pygame.K_5
    NUMBER_6 = pygame.K_6
    NUMBER_7 = pygame.K_7
    NUMBER_8 = pygame.K_8
    NUMBER_9 = pygame.K_9

    NUMPAD_0 = pygame.K_KP_0
    NUMPAD_1 = pygame.K_KP_1
    NUMPAD_2 = pygame.K_KP_2
    NUMPAD_3 = pygame.K_KP_3
    NUMPAD_4 = pygame.K_KP_4
    NUMPAD_5 = pygame.K_KP_5
    NUMPAD_6 = pygame.K_KP_6
    NUMPAD_7 = pygame.K_KP_7
    NUMPAD_8 = pygame.K_KP_8
    NUMPAD_9 = pygame.K_KP_9

    A = pygame.K_a
    B = pygame.K_b
    C = pygame.K_c
    D = pygame.K_d
    E = pygame.K_e
    F = pygame.K_f
    G = pygame.K_g
    H = pygame.K_h
    I = pygame.K_i
    J = pygame.K_j
    K = pygame.K_k
    L = pygame.K_l
    M = pygame.K_m
    N = pygame.K_n
    O = pygame.K_o
    P = pygame.K_p
    Q = pygame.K_q
    R = pygame.K_r
    S = pygame.K_s
    T = pygame.K_t
    U = pygame.K_u
    V = pygame.K_v
    W = pygame.K_w
    X = pygame.K_x
    Y = pygame.K_y
    Z = pygame.K_z



import struct
from pathlib import Path

import glm
import moderngl

import moderngl_window
from moderngl_window import geometry
from moderngl_window.opengl.projection import Projection3D
from moderngl_window.opengl.vao import VAO


class FragmentPicking(moderngl_window.WindowConfig):
    """
    Demonstrates how you can pick exact positions on a model
    efficiently using the gpu. We render the scene info
    to various layers in an offscreen framebuffer and
    use this to fetch the view position of the picked
    fragment. We can then apply the inverse modelview
    to create points positions that matches the original mesh.

    In this example we pick points on a mesh mapped with
    a texture containing heat information. We pick the
    position, heat value and the normal of the fragment.
    The normal is then used to hide points that point
    away from the camera.
    """

    title = "Fragment Picking"
    gl_version = 3, 3
    window_size = 1280, 720
    aspect_ratio = None
    resizable = True
    resource_dir = (Path(__file__) / "../../resources").resolve()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        print("window buffer size:", self.wnd.buffer_size)
        self.marker_file = Path("markers.bin")
        # Object rotation
        self.x_rot = 0
        self.y_rot = 0
        # Object position
        self.zoom = 0

        # Load scene cached to speed up loading!
        self.scene = self.load_scene("scenes/fragment_picking/centered.obj", cache=True)
        # Grab the raw mesh/vertex array
        self.mesh = self.scene.root_nodes[0].mesh.vao
        self.mesh_texture = self.scene.root_nodes[0].mesh.material.mat_texture.texture

        self.projection = Projection3D(
            fov=60,
            aspect_ratio=self.wnd.aspect_ratio,
            near=1.0,
            far=100.0,
        )

        # --- Offscreen render target
        # RGBA color/diffuse layer
        self.offscreen_diffuse = self.ctx.texture(self.wnd.buffer_size, 4)
        # Textures for storing normals (16 bit floats)
        self.offscreen_normals = self.ctx.texture(self.wnd.buffer_size, 4, dtype="f2")
        # Texture for storing the view positions rendered to framebuffer
        self.offscreen_viewpos = self.ctx.texture(self.wnd.buffer_size, 4, dtype="f4")
        # Texture for storing depth values
        self.offscreen_depth = self.ctx.depth_texture(self.wnd.buffer_size)
        # Create a framebuffer we can render to
        self.offscreen = self.ctx.framebuffer(
            color_attachments=[
                self.offscreen_diffuse,
                self.offscreen_normals,
                self.offscreen_viewpos,
            ],
            depth_attachment=self.offscreen_depth,
        )

        # This is just for temp changing depth texture parameters
        # temporary so we can use it as a normal texture
        self.depth_sampler = self.ctx.sampler(
            filter=(moderngl.LINEAR, moderngl.LINEAR),
            compare_func="",
        )

        # A fullscreen quad just for rendering offscreen textures to the window
        self.quad_fs = geometry.quad_fs()

        # --- Shaders
        # Simple program just rendering texture
        self.texture_program = self.load_program("programs/fragment_picking/texture.glsl")
        self.texture_program["texture0"].value = 0
        # Geomtry shader writing to two offscreen layers (color, normal) + depth
        self.geometry_program = self.load_program("programs/fragment_picking/geometry.glsl")
        self.geometry_program["texture0"].value = 0  # use texture channel 0

        # Shader for linearizing depth (debug visualization)
        self.linearize_depth_program = self.load_program("programs/linearize_depth.glsl")
        self.linearize_depth_program["texture0"].value = 0
        self.linearize_depth_program["near"].value = self.projection.near
        self.linearize_depth_program["far"].value = self.projection.far

        # Shader for picking the world position of a fragment
        self.fragment_picker_program = self.load_program("programs/fragment_picking/picker.glsl")
        self.fragment_picker_program["position_texture"].value = 0  # Read from texture channel 0
        self.fragment_picker_program["normal_texture"].value = 1  # Read from texture channel 1
        self.fragment_picker_program["diffuse_texture"].value = 2  # Read from texture channel 2

        # Picker geometry
        self.marker_byte_size = 7 * 4  # position + normal + temperature (7 x 32bit floats)
        self.picker_output = self.ctx.buffer(reserve=self.marker_byte_size)
        self.picker_vao = VAO(mode=moderngl.POINTS)

        # Shader for rendering markers
        self.marker_program = self.load_program("programs/fragment_picking/markers.glsl")
        self.marker_program["color"].value = 1.0, 0.0, 0.0, 1.0

        # Marker geometry
        self.marker_buffer = self.ctx.buffer(
            reserve=self.marker_byte_size * 1000
        )  # Reserve room for 1000 points
        self.marker_vao = VAO(name="markers", mode=moderngl.POINTS)
        self.marker_vao.buffer(
            self.marker_buffer, "3f 3f 1f", ["in_position", "in_normal", "temperature"]
        )
        self.num_markers = 0

        # Debug geometry
        self.quad_normals = geometry.quad_2d(size=(0.25, 0.25), pos=(0.75, 0.875))
        self.quad_depth = geometry.quad_2d(size=(0.25, 0.25), pos=(0.5, 0.875))
        self.quad_positions = geometry.quad_2d(size=(0.25, 0.25), pos=(0.25, 0.875))

    def on_render(self, time: float, frametime: float):
        self.ctx.enable(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        translation = glm.translate(glm.vec3(0, 0, -45 + self.zoom))
        rotation = glm.mat4(glm.quat(glm.vec3(self.y_rot, self.x_rot, 0)))
        self.modelview = translation * rotation

        # Render the scene to offscreen buffer
        self.offscreen.clear()
        self.offscreen.use()

        # Render the scene
        self.geometry_program["modelview"].write(self.modelview)
        self.geometry_program["projection"].write(self.projection.matrix)
        self.mesh_texture.use(location=0)  # bind texture from obj file to channel 0
        self.depth_sampler.use(location=0)
        self.mesh.render(self.geometry_program)  # render mesh
        self.depth_sampler.clear(location=0)

        # Activate the window as the render target
        self.ctx.screen.use()
        self.ctx.disable(moderngl.DEPTH_TEST)

        # Render offscreen diffuse layer to screen
        self.offscreen_diffuse.use(location=0)
        self.quad_fs.render(self.texture_program)

        # Render markers
        if self.num_markers > 0:
            self.ctx.point_size = 6.0  # Specify fragment size of the markers
            self.marker_program["modelview"].write(self.modelview)
            self.marker_program["projection"].write(self.projection.matrix)
            self.marker_vao.render(self.marker_program, vertices=self.num_markers)

        self.render_debug()

    def render_debug(self):
        """Debug rendering. Offscreen buffers"""
        # Debug rendering of normal and depth buffer
        self.offscreen_normals.use()
        self.quad_normals.render(self.texture_program)

        self.offscreen_depth.use(location=0)  # bind depth sampler to channel 0
        self.depth_sampler.use(location=0)  # temp override the parameters
        self.quad_depth.render(self.linearize_depth_program)
        self.depth_sampler.clear(location=0)  # Remove the override

        self.offscreen_viewpos.use()
        self.quad_positions.render(self.texture_program)

    def on_mouse_drag_event(self, x, y, dx, dy):
        """Pick up mouse drag movements"""
        self.x_rot -= dx / 100
        self.y_rot -= dy / 100

    def on_mouse_press_event(self, x, y, button):
        """Attempts to get the view position from a fragment"""

        # only care about right mouse button clicks
        if button != self.wnd.mouse.right:
            return

        # mouse coordinates starts in upper left corner
        # pixel positions starts and lower left corner
        pos = (
            int(x * self.wnd.pixel_ratio),
            int(self.wnd.buffer_height - (y * self.wnd.pixel_ratio)),
        )
        print("Picking mouse position", x, y)
        print("Viewport position", pos)

        self.fragment_picker_program["texel_pos"].value = pos
        self.fragment_picker_program["modelview"].write(self.modelview)
        self.offscreen_viewpos.use(location=0)
        self.offscreen_normals.use(location=1)
        self.offscreen_diffuse.use(location=2)
        self.picker_vao.transform(self.fragment_picker_program, self.picker_output, vertices=1)

        # Print position
        x, y, z, nx, ny, nz, temperature = struct.unpack("7f", self.picker_output.read())
        if z == 0.0:
            print("Point is not on the mesh")
            return

        print(f"Position: {x} {y} {z}")
        print(f"Normal: {nx} {ny} {nz}")
        print(f"Temperature: {round(temperature * 255)} (byte) {temperature} (float)")
        self.marker_buffer.write(
            self.picker_output.read(), offset=self.marker_byte_size * self.num_markers
        )
        self.num_markers += 1

    def on_mouse_scroll_event(self, x_offset, y_offset):
        self.zoom += y_offset

    def on_key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        # Key presses
        if action == keys.ACTION_PRESS:
            if key == keys.F1:
                print("Loading marker file")
                self.load_markers(self.marker_file)

            if key == keys.F2:
                print("Saving marker file")
                self.save_markers(self.marker_file)

    def load_markers(self, path: Path):
        """Loads markers from file"""
        if not self.marker_file.exists():
            return

        with open(self.marker_file, mode="rb") as fd:
            size = self.marker_file.stat().st_size
            self.num_markers = size // self.marker_byte_size
            self.marker_buffer.write(fd.read(), offset=0)

    def save_markers(self, path: Path):
        """Dump markers from file"""
        if self.num_markers == 0:
            return

        with open("markers.bin", mode="wb") as fd:
            fd.write(self.marker_buffer.read(size=self.num_markers * self.marker_byte_size))


if __name__ == "__main__":
    FragmentPicking.run()



"""
Custom usage: window creation and mapping callbacks functions at module level
"""

import math
import random

import moderngl_window
from moderngl_window.conf import settings
from moderngl_window.timers.clock import Timer

window = None


def main():
    global window
    # Configure to use pyglet window
    settings.WINDOW["class"] = "moderngl_window.context.pyglet.Window"
    window = moderngl_window.create_window_from_settings()

    # Map callback functions
    window.resize_func = on_resize
    window.iconify_func = on_iconify
    window.key_event_func = on_key_event
    window.mouse_position_event_func = on_mouse_position_event
    window.mouse_drag_event_func = on_mouse_drag_event
    window.mouse_scroll_event_func = on_mouse_scroll_event
    window.mouse_press_event_func = on_mouse_press_event
    window.mouse_release_event_func = on_mouse_release_event
    window.unicode_char_entered_func = on_unicode_char_entered

    timer = Timer()
    timer.start()

    while not window.is_closing:
        window.use()
        window.clear()

        time, frame_time = timer.next_frame()

        window.ctx.clear(
            (math.sin(time) + 1.0) / 2,
            (math.sin(time + 2) + 1.0) / 2,
            (math.sin(time + 3) + 1.0) / 2,
        )

        window.swap_buffers()

    window.destroy()


def on_render(time: float, frametime: float):
    # We can also check if a key is in press state here
    if window.is_key_pressed(window.keys.SPACE):
        print("User is holding SPACE button")


def on_resize(width: int, height: int):
    print("Window was resized. buffer size is {} x {}".format(width, height))


def on_iconify(iconify: bool):
    """Window hide/minimize and restore"""
    print("Window was iconified:", iconify)


def on_key_event(key, action, modifiers):
    keys = window.keys

    # Key presses
    if action == keys.ACTION_PRESS:
        if key == keys.SPACE:
            print("SPACE key was pressed")

        # Using modifiers (shift and ctrl)

        if key == keys.Z and modifiers.shift:
            print("Shift + Z was pressed")

        if key == keys.Z and modifiers.ctrl:
            print("ctrl + Z was pressed")

    # Key releases
    elif action == keys.ACTION_RELEASE:
        if key == keys.SPACE:
            print("SPACE key was released")

    # Move the window around with AWSD
    if action == keys.ACTION_PRESS:
        if key == keys.A:
            window.position = window.position[0] - 10, window.position[1]
        if key == keys.D:
            window.position = window.position[0] + 10, window.position[1]
        if key == keys.W:
            window.position = window.position[0], window.position[1] - 10
        if key == keys.S:
            window.position = window.position[0], window.position[1] + 10

        # toggle cursor
        if key == keys.C:
            window.cursor = not window.cursor

        # Shuffle window tittle
        if key == keys.T:
            title = list(window.title)
            random.shuffle(title)
            window.title = "".join(title)

        # Toggle mouse exclusivity
        if key == keys.M:
            window.mouse_exclusivity = not window.mouse_exclusivity


def on_mouse_position_event(x, y, dx, dy):
    print("Mouse position pos={} {} delta={} {}".format(x, y, dx, dy))


def on_mouse_drag_event(x, y, dx, dy):
    print("Mouse drag pos={} {} delta={} {}".format(x, y, dx, dy))


def on_mouse_scroll_event(x_offset, y_offset):
    print("mouse_scroll_event", x_offset, y_offset)


def on_mouse_press_event(x, y, button):
    print("Mouse button {} pressed at {}, {}".format(button, x, y))
    print("Mouse states:", window.mouse_states)


def on_mouse_release_event(x: int, y: int, button: int):
    print("Mouse button {} released at {}, {}".format(button, x, y))
    print("Mouse states:", window.mouse_states)


def on_unicode_char_entered(char):
    print("unicode_char_entered:", char)


if __name__ == "__main__":
    main()



import moderngl_window
from moderngl_window.text.bitmapped import TextWriter2D


class App(moderngl_window.WindowConfig):
    title = "Text"
    aspect_ratio = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.writer_quit = TextWriter2D()
        self.writer_quit.text = "Quit?"

        self.mode = "normal"  # normal / quit
        self.block_close = True

    def on_render(self, time, frame_time):
        if self.mode == "quit":
            self.writer_quit.draw((240, 380), size=120)

    def on_key_event(self, key, action, modifiers):
        keys = self.wnd.keys
        if self.mode == "quit":
            if key == keys.Y:
                self.block_close = False
                self.wnd.close()
            elif key == keys.N:
                self.block_close = True
                self.mode = "normal"

    def on_close(self, *args):
        if self.block_close:
            self.mode = "quit"
            self.wnd.is_closing = False


# Force running glfw window
moderngl_window.run_window_config(App, args=("--window", "glfw"))



"""
Bag of settings values
"""

# pylint: disable = invalid-name
import importlib
import os
import pathlib
from collections.abc import Generator, Iterable
from pprint import pformat
from types import ModuleType as Module
from typing import Any, Optional, Union

from moderngl_window.conf import default
from moderngl_window.exceptions import ImproperlyConfigured

SETTINGS_ENV_VAR = "MODERNGL_WINDOW_SETTINGS_MODULE"


class Settings:
    """
    Bag of settings values. New attributes can be freely added runtime.
    Various apply* methods are supplied so the user have full control over how
    settings values are initialized. This is especially useful for more custom usage.
    And instance of the `Settings` class is created when the `conf` module is imported.

    Attribute names must currently be in upper case to be recognized.

    Some examples of usage::

        from moderngl_window.conf import settings

        # Mandatory settings values
        try:
            value = settings.VALUE
        except KeyError:
            raise ValueError("This settings value is required")

        # Fallback in code
        value = getattr(settings, 'VALUE', 'default_value')

        # Pretty printed string representation for easy inspection
        print(settings)
    """

    WINDOW: dict[str, Any] = dict()
    """
    Window/screen properties. Most importantly the ``class`` attribute
    decides what class should be used to handle the window.

    .. code:: python

        # Default values
        WINDOW = {
            "gl_version": (3, 3),
            "class": "moderngl_window.context.pyglet.Window",
            "size": (1280, 720),
            "aspect_ratio": 16 / 9,
            "fullscreen": False,
            "resizable": True,
            "title": "ModernGL Window",
            "vsync": True,
            "cursor": True,
            "samples": 0,
        }

    Other Properties:

    - ``gl_version``: The minimum required major/minor OpenGL version
    - ``size``: The window size to open.
    - ``aspect_ratio`` is the enforced aspect ratio of the viewport.
    - ``fullscreen``: True if you want to create a context in fullscreen mode
    - ``resizable``: If the window should be resizable. This only applies in
      windowed mode.
    - ``vsync``: Only render one frame per screen refresh
    - ``title``: The visible title on the window in windowed mode
    - ``cursor``: Should the mouse cursor be visible on the screen? Disabling
      this is also useful in windowed mode when controlling the camera on some
      platforms as moving the mouse outside the window can cause issues.
    - ``Samples``: Number if samples used in multisampling. Values above 1
      enables multisampling.

    The created window frame buffer will by default use:

    - RGBA8 (32 bit per pixel)
    - 24 bit depth buffer
    - Double buffering
    - color and depth buffer is cleared for every frame

    """
    SCREENSHOT_PATH: Optional[str] = None
    """
    Absolute path to the directory screenshots will be saved by the screenshot module.
    Screenshots will end up in the project root of not defined.
    If a path is configured, the directory will be auto-created.
    """
    # Finders
    PROGRAM_FINDERS: list[str] = []
    """
    Finder classes for locating programs/shaders.

    .. code:: python

        # Default values
        PROGRAM_FINDERS = [
            "moderngl_window.finders.program.FileSystemFinder",
        ]
    """
    TEXTURE_FINDERS: list[str] = []
    """
    Finder classes for locating textures.

    .. code:: python

        # Default values
        TEXTURE_FINDERS = [
            "moderngl_window.finders.texture.FileSystemFinder",
        ]
    """
    SCENE_FINDERS: list[str] = []
    """
    Finder classes for locating scenes.

    .. code:: python

        # Default values
        SCENE_FINDERS = [
            "moderngl_window.finders.scene.FileSystemFinder",
        ]

    """
    DATA_FINDERS: list[str] = []
    """
    Finder classes for locating data files.

    .. code:: python

        # Default values
        DATA_FINDERS = [
            "moderngl_window.finders.data.FileSystemFinder",
        ]
    """
    # Finder dirs
    PROGRAM_DIRS: list[Union[str, pathlib.Path]] = []
    """
    Lists of `str` or `pathlib.Path` used by ``FileSystemFinder``
    to looks for programs/shaders.
    """
    TEXTURE_DIRS: list[Union[str, pathlib.Path]] = []
    """
    Lists of `str` or `pathlib.Path` used by ``FileSystemFinder``
    to looks for textures.
    """
    SCENE_DIRS: list[Union[str, pathlib.Path]] = []
    """
    Lists of `str` or `pathlib.Path` used by ``FileSystemFinder``
    to looks for scenes (obj, gltf, stl etc).
    """
    DATA_DIRS: list[Union[str, pathlib.Path]] = []
    """
    Lists of `str` or `pathlib.Path` used by ``FileSystemFinder``
    to looks for data files.
    """

    # Loaders
    PROGRAM_LOADERS: list[str] = []
    """
    Classes responsible for loading programs/shaders.

    .. code:: python

        # Default values
        PROGRAM_LOADERS = [
            'moderngl_window.loaders.program.single.Loader',
            'moderngl_window.loaders.program.separate.Loader',
        ]
    """
    TEXTURE_LOADERS: list[str] = []
    """
    Classes responsible for loading textures.

    .. code:: python

        # Default values
        TEXTURE_LOADERS = [
            'moderngl_window.loaders.texture.t2d.Loader',
            'moderngl_window.loaders.texture.array.Loader',
        ]
    """
    SCENE_LOADERS: list[str] = []
    """
    Classes responsible for loading scenes.

    .. code:: python

        # Default values
        SCENE_LOADERS = [
            "moderngl_window.loaders.scene.gltf.GLTF2",
            "moderngl_window.loaders.scene.wavefront.ObjLoader",
            "moderngl_window.loaders.scene.stl_loader.STLLoader",
        ]

    """
    DATA_LOADERS: list[str] = []
    """
    Classes responsible for loading data files.

    .. code:: python

        # Default values
        DATA_LOADERS = [
            'moderngl_window.loaders.data.binary.Loader',
            'moderngl_window.loaders.data.text.Loader',
            'moderngl_window.loaders.data.json.Loader',
        ]
    """

    def __init__(self) -> None:
        """Initialize settings with default values"""
        self.apply_default_settings()

    def apply_default_settings(self) -> None:
        """
        Apply keys and values from the default settings module
        located in this package. This is to ensure we always
        have the minimal settings for the system to run.

        If replacing or customizing the settings class
        you must always apply default settings to ensure
        compatibility when new settings are added.
        """
        self.apply_from_module(default)

    def apply_settings_from_env(self) -> None:
        """
        Apply settings from ``MODERNGL_WINDOW_SETTINGS_MODULE`` environment variable.
        If the environment variable is undefined no action will be taken.
        Normally this would be used to easily be able to switch between
        different configuration by setting env vars before executing the program.

        Example::

            import os
            from moderngl_window.conf import settings

            os.environ['MODERNGL_WINDOW_SETTINGS_MODULE'] = 'python.path.to.module'
            settings.apply_settings_from_env()

        Raises:
            ImproperlyConfigured if the module was not found
        """
        name = os.environ.get(SETTINGS_ENV_VAR)
        if name:
            self.apply_from_module_name(name)

    def apply_from_module_name(self, settings_module_name: str) -> None:
        """
        Apply settings from a python module by supplying the full
        pythonpath to the module.

        Args:
            settings_module_name (str): Full python path to the module

        Raises:
            ImproperlyConfigured if the module was not found
        """
        try:
            module = importlib.import_module(settings_module_name)
        except ImportError as ex:
            raise ImproperlyConfigured(
                "Settings module '{}' not found. From importlib: {}".format(
                    settings_module_name,
                    ex,
                )
            )

        self.apply_from_module(module)

    def apply_from_dict(self, data: dict[str, Any]) -> None:
        """
        Apply settings values from a dictionary

        Example::

            >> from moderngl_window.conf import settings
            >> settings.apply_dict({'SOME_VALUE': 1})
            >> settings.SOME_VALUE
            1
        """
        self.apply_from_iterable(data.items())

    def apply_from_module(self, module: Module) -> None:
        """
        Apply settings values from a python module

        Example::

            my_settings.py module containing the following line:
            SOME_VALUE = 1

            >> from moderngl_window.conf import settings
            >> import my_settings
            >> settings.apply_module(my_settings)
            >> settings.SOME_VALUE
            1
        """
        self.apply_from_iterable(module.__dict__.items())

    def apply_from_cls(self, cls: Any) -> None:
        """
        Apply settings values from a class namespace

        Example::

            >> from moderngl_window.conf import settings
            >> class MySettings:
            >>    SOME_VALUE = 1
            >>
            >> settings.apply(MySettings)
            >> settings.SOME_VALUE
            1
        """
        self.apply_from_iterable(cls.__dict__.items())

    def apply_from_iterable(self, iterable: Iterable[tuple[str, Any]]) -> None:
        """
        Apply (key, value) pairs from an iterable or generator
        """
        if not isinstance(iterable, Iterable) and not isinstance(self, Generator):
            raise ValueError(
                "Input value is not a generator or iterable, but of type: {}".format(type(iterable))
            )

        for name, value in iterable:
            if name.isupper():
                setattr(self, name, value)

    def to_dict(self) -> dict[str, Any]:
        """Create a dict representation of the settings
        Only uppercase attributes are included

        Returns:
            dict: dict representation
        """
        return {k: v for k, v in self.__dict__.items() if k.upper()}

    def __repr__(self) -> str:
        return "\n".join(
            "{}={}".format(k, pformat(v, indent=2)) for k, v in self.__dict__.items() if k.isupper()
        )


settings = Settings()



"""
Example showing how we can modify the default cli argument parser.

Imagine we're making a small utility displaying a 3D object
from some file. We require a path to the file and and optional
argument if the model should be rendered as wireframe or not.
In addition we have an optional argument for changing the
window title.

Possible arguments should be:

python script.py path/to/file
python script.py path/to/file --wireframe
python script.py path/to/file --wireframe --title "Custom Window Title"
python script.py path/to/file --title "Custom Window Title"

"""

import math

import moderngl_window


class ModifyParser(moderngl_window.WindowConfig):
    title = "Modify Parser"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        print("all arguments :", self.argv)
        # Print our custom argument
        print("path      :", self.argv.path)
        print("wireframe :", self.argv.wireframe)
        print("title     :", self.argv.title)

        if self.argv.title:
            self.wnd.title = self.argv.title

    @classmethod
    def add_arguments(cls, parser):
        # Mandatory positional argument for the file to load
        parser.add_argument(
            "path",
            help="Path to the model to display",
        )
        # Optional flag for rendering the model in wireframe
        # This is simply enabled by adding "--wireframe" with no arguments
        parser.add_argument(
            "--wireframe",
            action="store_true",
            default=False,
            help="Display the model as a wireframe",
        )
        # Optional argument for window title
        parser.add_argument(
            "--title",
            type=str,
            help="Override the window title",
        )

    def on_render(self, time, frame_time):
        # Placeholder content
        self.ctx.clear(
            (math.sin(time) + 1.0) / 2,
            (math.sin(time + 2) + 1.0) / 2,
            (math.sin(time + 3) + 1.0) / 2,
        )


if __name__ == "__main__":
    ModifyParser.run()



"""
Press and drag the mouse to scroll the texture
"""

from pathlib import Path

import moderngl_window
from moderngl_window import geometry, resources

resources.register_dir((Path(__file__).parent / "resources").resolve())


class QuadFullscreenScroll(moderngl_window.WindowConfig):
    title = "Quad Fullscreen Scroll"
    aspect_ratio = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.quad = geometry.quad_fs()
        self.texture = self.load_texture_2d("textures/python-bg.png")
        self.prog = self.load_program("programs/texture_mouse_scroll.glsl")
        self.mouse_pos = 0, 0

    def on_render(self, time: float, frame_time: float):
        self.ctx.clear()

        self.texture.use(location=0)
        self.prog["texture0"] = 0
        self.prog["offset"] = (
            -self.mouse_pos[0] / self.wnd.buffer_width,
            self.mouse_pos[1] / self.wnd.buffer_height,
        )
        self.quad.render(self.prog)

    def on_mouse_drag_event(self, x, y, dx, dy):
        self.mouse_pos = self.mouse_pos[0] + dx, self.mouse_pos[1] + dy
        print(self.mouse_pos)


if __name__ == "__main__":
    QuadFullscreenScroll.run()



import math

import moderngl_window


class BasicWindowConfig(moderngl_window.WindowConfig):
    """Minimal WindowConfig example"""

    gl_version = (3, 3)
    title = "Basic Window Config"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_render(self, time: float, frametime: float):
        self.ctx.clear(
            (math.sin(time) + 1.0) / 2,
            (math.sin(time + 2) + 1.0) / 2,
            (math.sin(time + 3) + 1.0) / 2,
        )


if __name__ == "__main__":
    BasicWindowConfig.run()



import json
import logging
from typing import Any

from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.loaders.base import BaseLoader

logger = logging.getLogger(__name__)


class Loader(BaseLoader):
    kind = "json"
    file_extensions = [
        [".json"],
    ]

    def load(self) -> dict[Any, Any]:
        """Load a file as json

        Returns:
            dict: The json contents
        """
        assert self.meta.path is not None, "the path is empty for this loader"
        self.meta.resolved_path = self.find_data(self.meta.path)

        if not self.meta.resolved_path:
            raise ImproperlyConfigured("Data file '{}' not found".format(self.meta.path))

        logger.info("Loading: %s", self.meta.path)

        with open(str(self.meta.resolved_path), "r") as fd:
            return json.loads(fd.read())



# flake8: noqa E741
import platform

import pyglet

# On OS X we need to disable the shadow context
# because the 2.1 shadow context cannot be upgrade to a 3.3+ core
if platform.system() == "Darwin":
    pyglet.options["shadow_window"] = False

pyglet.options["debug_gl"] = False

from pyglet.window import key

from moderngl_window.context.base import BaseKeys


class Keys(BaseKeys):
    """
    Namespace mapping pyglet specific key constants
    """

    ESCAPE = key.ESCAPE
    SPACE = key.SPACE
    ENTER = key.ENTER
    PAGE_UP = key.PAGEUP
    PAGE_DOWN = key.PAGEDOWN
    LEFT = key.LEFT
    RIGHT = key.RIGHT
    UP = key.UP
    DOWN = key.DOWN
    LEFT_SHIFT = key.LSHIFT
    RIGHT_SHIFT = key.RSHIFT
    LEFT_CTRL = key.LCTRL

    TAB = key.TAB
    COMMA = key.COMMA
    MINUS = key.MINUS
    PERIOD = key.PERIOD
    SLASH = key.SLASH
    SEMICOLON = key.SEMICOLON
    EQUAL = key.EQUAL
    LEFT_BRACKET = key.BRACELEFT
    RIGHT_BRACKET = key.BRACERIGHT
    BACKSLASH = key.BACKSLASH
    BACKSPACE = key.BACKSPACE
    INSERT = key.INSERT
    DELETE = key.DELETE
    HOME = key.HOME
    END = key.END
    CAPS_LOCK = key.CAPSLOCK

    F1 = key.F1
    F2 = key.F2
    F3 = key.F3
    F4 = key.F4
    F5 = key.F5
    F6 = key.F6
    F7 = key.F7
    F8 = key.F8
    F9 = key.F9
    F10 = key.F10
    F11 = key.F11
    F12 = key.F12

    NUMBER_0 = key._0
    NUMBER_1 = key._1
    NUMBER_2 = key._2
    NUMBER_3 = key._3
    NUMBER_4 = key._4
    NUMBER_5 = key._5
    NUMBER_6 = key._6
    NUMBER_7 = key._7
    NUMBER_8 = key._8
    NUMBER_9 = key._9

    NUMPAD_0 = key.NUM_0
    NUMPAD_1 = key.NUM_1
    NUMPAD_2 = key.NUM_2
    NUMPAD_3 = key.NUM_3
    NUMPAD_4 = key.NUM_4
    NUMPAD_5 = key.NUM_5
    NUMPAD_6 = key.NUM_6
    NUMPAD_7 = key.NUM_7
    NUMPAD_8 = key.NUM_8
    NUMPAD_9 = key.NUM_9

    A = key.A
    B = key.B
    C = key.C
    D = key.D
    E = key.E
    F = key.F
    G = key.G
    H = key.H
    I = key.I
    J = key.J
    K = key.K
    L = key.L
    M = key.M
    N = key.N
    O = key.O
    P = key.P
    Q = key.Q
    R = key.R
    S = key.S
    T = key.T
    U = key.U
    V = key.V
    W = key.W
    X = key.X
    Y = key.Y
    Z = key.Z



from pathlib import Path

import glm
import moderngl
from base import CameraWindow

import moderngl_window


class CubeModel(CameraWindow):
    aspect_ratio = 16 / 9
    resource_dir = Path(__file__).parent.resolve() / "resources"
    title = "Cube Model"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True

        self.scene = self.load_scene("scenes/crate.obj")
        # self.scene = self.load_scene('scenes/Apollo_17.stl')

        self.camera.projection.update(near=0.1, far=100.0)
        self.camera.velocity = 7.0
        self.camera.mouse_sensitivity = 0.3

    def on_render(self, time: float, frametime: float):
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        translation = glm.translate(glm.vec3(0, 0, -1.5))
        rotation = glm.mat4(glm.quat(glm.vec3(0, 0, 0)))
        model_matrix = translation * rotation
        camera_matrix = self.camera.matrix * model_matrix

        self.scene.draw(
            projection_matrix=self.camera.projection.matrix,
            camera_matrix=camera_matrix,
            time=time,
        )


if __name__ == "__main__":
    moderngl_window.run_window_config(CubeModel)



"""
Shadow mapping example from:
https://www.opengl-tutorial.org/intermediate-tutorials/tutorial-16-shadow-mapping/
"""

import math
from pathlib import Path

import glm
import moderngl
from base import CameraWindow

from moderngl_window import geometry


class ShadowMapping(CameraWindow):
    title = "Shadow Mapping"
    resource_dir = (Path(__file__) / "../../resources").resolve()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.camera.projection.update(near=1, far=200)
        self.wnd.mouse_exclusivity = True

        # Offscreen buffer
        offscreen_size = 1024, 1024
        self.offscreen_depth = self.ctx.depth_texture(offscreen_size)
        self.offscreen_depth.compare_func = ""
        self.offscreen_depth.repeat_x = False
        self.offscreen_depth.repeat_y = False
        self.offscreen_color = self.ctx.texture(offscreen_size, 4)

        self.offscreen = self.ctx.framebuffer(
            color_attachments=[self.offscreen_color],
            depth_attachment=self.offscreen_depth,
        )

        # Scene geometry
        self.floor = geometry.cube(size=(25.0, 1.0, 25.0))
        self.wall = geometry.cube(size=(1.0, 5, 25), center=(-12.5, 2, 0))
        self.sphere = geometry.sphere(radius=5.0, sectors=64, rings=32)
        self.sun = geometry.sphere(radius=1.0)

        # Debug geometry
        self.offscreen_quad = geometry.quad_2d(size=(0.5, 0.5), pos=(0.75, 0.75))
        self.offscreen_quad2 = geometry.quad_2d(size=(0.5, 0.5), pos=(0.25, 0.75))

        # Programs
        self.raw_depth_prog = self.load_program("programs/shadow_mapping/raw_depth.glsl")
        self.basic_light = self.load_program("programs/shadow_mapping/directional_light.glsl")
        self.basic_light["shadowMap"].value = 0
        self.basic_light["color"].value = 1.0, 1.0, 1.0, 1.0
        self.shadowmap_program = self.load_program("programs/shadow_mapping/shadowmap.glsl")
        self.texture_prog = self.load_program("programs/texture.glsl")
        self.texture_prog["texture0"].value = 0
        self.sun_prog = self.load_program("programs/cube_simple.glsl")
        self.sun_prog["color"].value = 1, 1, 0, 1
        self.lightpos = 0, 0, 0

    def on_render(self, time: float, frametime: float):
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)
        self.lightpos = glm.vec3(math.sin(time) * 20, 5, math.cos(time) * 20)
        scene_pos = glm.vec3(0, -5, -32)

        # --- PASS 1: Render shadow map
        self.offscreen.clear()
        self.offscreen.use()

        depth_projection = glm.ortho(-20, 20, -20, 20, -20, 40)
        depth_view = glm.lookAt(self.lightpos, (0, 0, 0), (0, 1, 0))
        depth_mvp = depth_projection * depth_view
        self.shadowmap_program["mvp"].write(depth_mvp)

        self.floor.render(self.shadowmap_program)
        self.wall.render(self.shadowmap_program)
        self.sphere.render(self.shadowmap_program)

        # --- PASS 2: Render scene to screen
        self.wnd.use()
        self.basic_light["m_proj"].write(self.camera.projection.matrix)
        self.basic_light["m_camera"].write(self.camera.matrix)
        self.basic_light["m_model"].write(glm.translate(glm.vec3(scene_pos)))
        bias_matrix = glm.mat4(
            [
                [0.5, 0.0, 0.0, 0.0],
                [0.0, 0.5, 0.0, 0.0],
                [0.0, 0.0, 0.5, 0.0],
                [0.5, 0.5, 0.5, 1.0],
            ],
        )
        self.basic_light["m_shadow_bias"].write(bias_matrix * depth_mvp)
        self.basic_light["lightDir"].write(self.lightpos)
        self.offscreen_depth.use(location=0)
        self.floor.render(self.basic_light)
        self.wall.render(self.basic_light)
        self.sphere.render(self.basic_light)

        # Render the sun position
        self.sun_prog["m_proj"].write(self.camera.projection.matrix)
        self.sun_prog["m_camera"].write(self.camera.matrix)
        self.sun_prog["m_model"].write(glm.translate(glm.vec3(self.lightpos + scene_pos)))
        self.sun.render(self.sun_prog)

        # --- PASS 3: Debug ---
        # self.ctx.enable_only(moderngl.NOTHING)
        self.offscreen_depth.use(location=0)
        self.offscreen_quad.render(self.raw_depth_prog)
        # self.offscreen_color.use(location=0)
        # self.offscreen_quad2.render(self.texture_prog)


if __name__ == "__main__":
    ShadowMapping.run()



from typing import Any, Optional, Union

import moderngl
import numpy
import numpy.typing as npt

import moderngl_window as mglw
from moderngl_window.opengl import types

# For sanity checking draw modes when creating the VAO
DRAW_MODES = {
    moderngl.TRIANGLES: "TRIANGLES",
    moderngl.TRIANGLE_FAN: "TRIANGLE_FAN",
    moderngl.TRIANGLE_STRIP: "TRIANGLE_STRIP",
    moderngl.TRIANGLES_ADJACENCY: "TRIANGLES_ADJACENCY",
    moderngl.TRIANGLE_STRIP_ADJACENCY: "TRIANGLE_STRIP_ADJACENCY",
    moderngl.POINTS: "POINTS",
    moderngl.LINES: "LINES",
    moderngl.LINE_STRIP: "LINE_STRIP",
    moderngl.LINE_LOOP: "LINE_LOOP",
    moderngl.LINES_ADJACENCY: "LINES_ADJACENCY",
}


class BufferInfo:
    """Container for a vbo with additional information"""

    def __init__(
        self,
        buffer: moderngl.Buffer,
        buffer_format: str,
        attributes: list[str] = [],
        per_instance: bool = False,
    ):
        """
        :param buffer: The vbo object
        :param format: The format of the buffer
        """
        self.buffer = buffer
        self.attrib_formats = types.parse_attribute_formats(buffer_format)
        self.attributes = attributes
        self.per_instance = per_instance

        # Sanity check byte size
        if self.buffer.size % self.vertex_size != 0:
            raise VAOError(
                "Buffer with type {} has size not aligning with {}. Remainder: {}".format(
                    buffer_format, self.vertex_size, self.buffer.size % self.vertex_size
                )
            )

        self.vertices = self.buffer.size // self.vertex_size

    @property
    def vertex_size(self) -> int:
        return sum(f.bytes_total for f in self.attrib_formats)

    def content(self, attributes: list[str]) -> Optional[tuple[object, ...]]:
        """Build content tuple for the buffer

        Returns
            The first value is the moderngl buffer
            From the third to the end, it is the attributes of the class"""
        formats = []
        attrs = []
        for attrib_format, attrib in zip(self.attrib_formats, self.attributes):
            if attrib not in attributes:
                formats.append(attrib_format.pad_str())
                continue

            formats.append(attrib_format.format)
            attrs.append(attrib)

            attributes.remove(attrib)

        if len(attrs) == 0:
            return None

        return (
            self.buffer,
            "{}{}".format(" ".join(formats), "/i" if self.per_instance else ""),
            *attrs,
        )

    def has_attribute(self, name: str) -> bool:
        return name in self.attributes


class VAO:
    """
    Represents a vertex array object.

    This is a wrapper class over ``moderngl.VertexArray`` to make interactions
    with programs/shaders simpler. Named buffers are added corresponding with
    attribute names in a vertex shader. When rendering the VAO an internal
    ``moderngl.VertextArray`` is created automatically mapping the named buffers
    compatible with the supplied program. This program is cached internally.

    The shader program doesn't need to use all the buffers registered in
    this wrapper. When a subset is used only the used buffers are mapped
    and the appropriate padding is calculated when interleaved data is used.

    You are not required to use this class, but most methods in the
    system creating vertexbuffers will return this type. You can obtain
    a single ``moderngl.VertexBuffer`` instance by calling :py:meth:`VAO.instance`
    method if you prefer to work directly on moderngl instances.

    Example::

        # Separate buffers
        vao = VAO(name="test", mode=moderngl.POINTS)
        vao.buffer(positions, '3f', ['in_position'])
        vao.buffer(velocities, '3f', ['in_velocities'])

        # Interleaved
        vao = VAO(name="test", mode=moderngl.POINTS)
        vao.buffer(interleaved_data, '3f 3f', ['in_position', 'in_velocities'])

    .. code:: glsl

        # GLSL vertex shader in attributes
        in vec3 in_position;
        in vec3 in_velocities;

    """

    def __init__(self, name: str = "", mode: int = moderngl.TRIANGLES):
        """Create and empty VAO with a name and default render mode.

        Example::

            VAO(name="cube", mode=moderngl.TRIANGLES)

        Keyword Args:
            name (str): Optional name for debug purposes
            mode (int): Default draw mode
        """
        self.name = name
        self.mode = mode

        try:
            DRAW_MODES[self.mode]
        except KeyError:
            raise VAOError("Invalid draw mode. Options are {}".format(DRAW_MODES.values()))

        self._buffers: list[BufferInfo] = []
        self._index_buffer: Optional[moderngl.Buffer] = None
        self._index_element_size: Optional[int] = None

        self.vertex_count = 0
        self.vaos: dict[Any, moderngl.VertexArray] = {}

    @property
    def ctx(self) -> moderngl.Context:
        """moderngl.Context: The active moderngl context"""
        return mglw.ctx()

    def render(
        self,
        program: moderngl.Program,
        mode: Optional[int] = None,
        vertices: int = -1,
        first: int = 0,
        instances: int = 1,
    ) -> None:
        """Render the VAO.

        An internal ``moderngl.VertexBuffer`` with compatible buffer bindings
        is automatically created on the fly and cached internally.

        Args:
            program: The ``moderngl.Program``
        Keyword Args:
            mode: Override the draw mode (``TRIANGLES`` etc)
            vertices (int): The number of vertices to transform
            first (int): The index of the first vertex to start with
            instances (int): The number of instances
        """
        vao = self.instance(program)

        if mode is None:
            mode = self.mode

        vao.render(mode, vertices=vertices, first=first, instances=instances)

    def render_indirect(
        self,
        program: moderngl.Program,
        buffer: moderngl.Buffer,
        mode: Optional[int] = None,
        count: int = -1,
        *,
        first: int = 0,
    ) -> None:
        """
        The render primitive (mode) must be the same as the input primitive of the
        GeometryShader.

        The draw commands are 5 integers:
        (count, instanceCount, firstIndex, baseVertex, baseInstance).

        Args:
            program: The ``moderngl.Program``
            buffer: The ``moderngl.Buffer`` containing indirect draw commands
        Keyword Args:
            mode (int): By default :py:data:`TRIANGLES` will be used.
            count (int): The number of draws.
            first (int): The index of the first indirect draw command.
        """
        vao = self.instance(program)

        if mode is None:
            mode = self.mode

        vao.render_indirect(buffer, mode=mode, count=count, first=first)

    def transform(
        self,
        program: moderngl.Program,
        buffer: moderngl.Buffer,
        mode: Optional[int] = None,
        vertices: int = -1,
        first: int = 0,
        instances: int = 1,
    ) -> None:
        """Transform vertices. Stores the output in a single buffer.

        Args:
            program: The ``moderngl.Program``
            buffer: The ``moderngl.buffer`` to store the output
        Keyword Args:
            mode: Draw mode (for example ``moderngl.POINTS``)
            vertices (int): The number of vertices to transform
            first (int): The index of the first vertex to start with
            instances (int): The number of instances
        """
        vao = self.instance(program)

        if mode is None:
            mode = self.mode

        vao.transform(buffer, mode=mode, vertices=vertices, first=first, instances=instances)

    def buffer(
        self,
        buffer: Union[moderngl.Buffer, npt.NDArray[Any], bytes],
        buffer_format: str,
        attribute_names: Union[list[str], str],
    ) -> moderngl.Buffer:
        """Register a buffer/vbo for the VAO. This can be called multiple times.
        adding multiple buffers (interleaved or not).

        Args:
            buffer:
                The buffer data. Can be ``numpy.array``, ``moderngl.Buffer`` or ``bytes``.
            buffer_format (str):
                The format of the buffer. (eg. ``3f 3f`` for interleaved positions and normals).
            attribute_names:
                A list of attribute names this buffer should map to.
        Returns:
            The ``moderngl.Buffer`` instance object. This is handy when providing ``bytes``
            and ``numpy.array``.
        """
        if not isinstance(attribute_names, list):
            attribute_names = [
                attribute_names,
            ]

        if type(buffer) not in [moderngl.Buffer, numpy.ndarray, bytes]:
            raise VAOError(
                (
                    "buffer parameter must be a moderngl.Buffer, numpy.ndarray or bytes instance"
                    "(not {})".format(type(buffer))
                )
            )

        if isinstance(buffer, numpy.ndarray):
            buffer = self.ctx.buffer(buffer.tobytes())

        if isinstance(buffer, bytes):
            buffer = self.ctx.buffer(data=buffer)

        formats = buffer_format.split()
        if len(formats) != len(attribute_names):
            raise VAOError(
                "Format '{}' does not describe attributes {}".format(buffer_format, attribute_names)
            )

        self._buffers.append(BufferInfo(buffer, buffer_format, attribute_names))
        self.vertex_count = self._buffers[-1].vertices

        return buffer

    def index_buffer(
        self, buffer: Union[moderngl.Buffer, npt.NDArray[Any], bytes], index_element_size: int = 4
    ) -> None:
        """Set the index buffer for this VAO.

        Args:
            buffer: ``moderngl.Buffer``, ``numpy.array`` or ``bytes``
        Keyword Args:
            index_element_size (int): Byte size of each element. 1, 2 or 4
        """
        if type(buffer) not in [moderngl.Buffer, numpy.ndarray, bytes]:
            raise VAOError(
                "buffer parameter must be a moderngl.Buffer, numpy.ndarray or bytes instance"
            )

        if isinstance(buffer, numpy.ndarray):
            buffer = self.ctx.buffer(buffer.tobytes())

        if isinstance(buffer, bytes):
            buffer = self.ctx.buffer(data=buffer)

        self._index_buffer = buffer
        self._index_element_size = index_element_size

    def instance(self, program: moderngl.Program) -> moderngl.VertexArray:
        """Obtain the ``moderngl.VertexArray`` instance for the program.

        The instance is only created once and cached internally.

        Args:
            program (moderngl.Program): The program

        Returns:
            ``moderngl.VertexArray``: instance
        """
        vao = self.vaos.get(program.glo)
        if vao is not None and isinstance(vao, moderngl.VertexArray):
            return vao

        program_attributes = [
            name
            for name, attr in program._members.items()
            if isinstance(attr, moderngl.Attribute) and not attr.name.startswith("gl_")
        ]

        # Make sure all attributes are covered
        for attrib_name in program_attributes:
            # Do we have a buffer mapping to this attribute?
            if not sum(buffer.has_attribute(attrib_name) for buffer in self._buffers):
                raise VAOError(
                    (
                        "VAO {} doesn't have attribute {} for program {}.\n"
                        "Program attributes: {}.\n"
                        "VAO attributes: {}"
                    ).format(
                        self.name,
                        attrib_name,
                        program,
                        program_attributes,
                        [attr for buff in self._buffers for attr in buff.attributes],
                    )
                )

        vao_content = []

        # Pick out the attributes we can actually map
        for buffer in self._buffers:
            content = buffer.content(program_attributes)
            if content:
                vao_content.append(content)

        # Any attribute left is not accounted for
        if program_attributes:
            raise VAOError(
                "Did not find a buffer mapping for {}".format([n for n in program_attributes])
            )

        # Create the vao
        if self._index_buffer:
            vao = self.ctx.vertex_array(
                program,
                vao_content,
                self._index_buffer,
                self._index_element_size,
            )
        else:
            vao = self.ctx.vertex_array(program, vao_content)

        self.vaos[program.glo] = vao
        return vao

    def release(self, buffer: bool = True) -> None:
        """Destroy all internally cached vaos and release all buffers.

        Keyword Args:
            buffers (bool): also release buffers
        """
        for _, vao in self.vaos.items():
            vao.release()

        self.vaos = {}

        if buffer:
            for buff in self._buffers:
                buff.buffer.release()

            if self._index_buffer:
                self._index_buffer.release()

        self._buffers = []

    def get_buffer_by_name(self, name: str) -> Optional[BufferInfo]:
        """Get the BufferInfo associated with a specific attribute name

        If no buffer is associated with the name `None` will be returned.

        Args:
            name (str): Name of the mapped attribute
        Returns:
            BufferInfo: BufferInfo instance
        """
        for buffer in self._buffers:
            if name in buffer.attributes:
                return buffer

        return None


class VAOError(Exception):
    pass



from typing import Any, Optional

from moderngl_window.geometry.attributes import AttributeNames
from moderngl_window.meta.base import ResourceDescription


class SceneDescription(ResourceDescription):
    """Describes a scene to load.

    The correct loader is resolved by looking at the file extension.
    This can be overridden by specifying a ``kind`` that maps directly
    to a specific loader class.

    .. code:: python

        # Wavefront/obj file
        SceneDescription(path='scenes/cube.obj')

        # stl file
        SceneDescription(path='scenes/crater.stl')

        # GLTF 2 file
        SceneDescription(path='scenes/sponza.gltf')

    The user can also override what buffer/attribute names
    should be used by specifying ``attr_names``.

    A ``cache`` option is also available as some scene loaders
    supports converting the file into a different format
    on the fly to speed up loading.
    """

    default_kind = ""
    resource_type = "scenes"

    def __init__(
        self,
        path: Optional[str] = None,
        kind: Optional[str] = None,
        cache: bool = False,
        attr_names: type[AttributeNames] = AttributeNames,
        **kwargs: Any,
    ):
        """Create a scene description.

        Keyword Args:
            path (str): Path to resource
            kind (str): Loader kind
            cache (str): Use the loader caching system if present
            attr_names (AttributeNames): Attrib name config
            **kwargs: Optional custom attributes
        """
        if attr_names is None:
            attr_names = AttributeNames

        kwargs.update({"path": path, "kind": kind, "cache": cache, "attr_names": attr_names})
        super().__init__(**kwargs)

    @property
    def cache(self) -> bool:
        """bool: Use cache feature in scene loader"""
        return bool(self._kwargs["cache"])

    @property
    def attr_names(self) -> AttributeNames:
        """AttributeNames: Attribute name config"""
        return self._kwargs["attr_names"]



"""
Shader Registry
"""

from typing import Union

import moderngl

from moderngl_window.meta import ResourceDescription, TextureDescription
from moderngl_window.resources.base import BaseRegistry

TextureAny = Union[
    moderngl.Texture,
    moderngl.TextureArray,
    moderngl.TextureCube,
    moderngl.Texture3D,
]


class Textures(BaseRegistry):
    """Handles texture resources"""

    settings_attr = "TEXTURE_LOADERS"
    meta: TextureDescription

    def load(self, meta: ResourceDescription) -> TextureAny:
        """Loads a texture with the configured loaders.

        Args:
            meta (:py:class:`~moderngl_window.meta.texture.TextureDescription`):
            The resource description
        Returns:
            moderngl.Texture: 2d texture
        Returns:
            moderngl.TextureArray: texture array if ``layers`` is supplied
        """
        texture = super().load(meta)
        assert (
            isinstance(texture, moderngl.Texture)
            or isinstance(texture, moderngl.TextureArray)
            or isinstance(texture, moderngl.TextureCube)
            or isinstance(texture, moderngl.Texture3D)
        ), f"{meta} did not load a texture. Please correct it"
        return texture


textures = Textures()



from unittest import TestCase

from moderngl_window.meta import (DataDescription, ProgramDescription,
                                  SceneDescription, TextureDescription)
from moderngl_window.resources.base import ResourceDescription


class ResourceDescriptionTestCase(TestCase):
    path = 'path/to'
    label = 'label'
    kind = 'somekind'

    def test_base(self):
        """Create a base resource description"""
        instance = ResourceDescription(
            path=self.path,
            label=self.label,
            kind=self.kind,
        )
        self.inspect_base_properties(instance)
        str(instance)
        repr(instance)

    def inspect_base_properties(self, instance):
        self.assertEqual(instance.path, self.path)
        self.assertEqual(instance.label, self.label)
        self.assertEqual(instance.kind, self.kind)
        self.assertEqual(instance.loader_cls, None)
        self.assertEqual(instance.resolved_path, None)
        self.assertIsInstance(instance.attrs, dict)

        instance.kind = 'otherkind'
        instance.loader_cls = 'fake_loader_cls'
        self.assertEqual(instance.kind, 'otherkind')
        self.assertEqual(instance.loader_cls, 'fake_loader_cls')

    def test_data(self):
        """Create a DataDescription"""
        instance = DataDescription(
            path=self.path,
            kind=self.kind,
            label=self.label,
        )
        self.inspect_base_properties(instance)

    def test_program(self):
        """Create ProgramDescription"""
        instance = ProgramDescription(
            path=self.path,
            kind=self.kind,
            label=self.label,
            vertex_shader='vertex_shader',
            geometry_shader='geometry_shader',
            fragment_shader='fragment_shader',
            tess_control_shader='tess_control_shader',
            tess_evaluation_shader='tess_evaluation_shader',
        )
        self.inspect_base_properties(instance)
        self.assertEqual(instance.reloadable, False)
        self.assertEqual(instance.vertex_shader, 'vertex_shader')
        self.assertEqual(instance.geometry_shader, 'geometry_shader')
        self.assertEqual(instance.fragment_shader, 'fragment_shader')
        self.assertEqual(instance.tess_control_shader, 'tess_control_shader')
        self.assertEqual(instance.tess_evaluation_shader, 'tess_evaluation_shader')

    def test_scene(self):
        """Create SceneDescription"""
        instance = SceneDescription(
            path=self.path,
            kind=self.kind,
            label=self.label,
        )
        self.assertFalse(instance.cache)
        self.inspect_base_properties(instance)

    def test_texture(self):
        """Create TextureDescription"""
        instance = TextureDescription(
            path=self.path,
            kind=self.kind,
            label=self.label,
        )
        self.inspect_base_properties(instance)
        self.assertEqual(instance.flip_x, False)
        self.assertEqual(instance.flip_y, True)
        self.assertEqual(instance.mipmap, False)
        self.assertEqual(instance.mipmap_levels, None)
        self.assertEqual(instance.anisotropy, 1.0)



# flake8: noqa E741
import glfw

from moderngl_window.context.base import BaseKeys

GLFW_key = int


class Keys(BaseKeys):
    """
    Namespace defining glfw specific keys constants
    """

    ACTION_PRESS = glfw.PRESS
    ACTION_RELEASE = glfw.RELEASE

    ESCAPE = glfw.KEY_ESCAPE
    SPACE = glfw.KEY_SPACE
    ENTER = glfw.KEY_ENTER
    PAGE_UP = glfw.KEY_PAGE_UP
    PAGE_DOWN = glfw.KEY_PAGE_DOWN
    LEFT = glfw.KEY_LEFT
    RIGHT = glfw.KEY_RIGHT
    UP = glfw.KEY_UP
    DOWN = glfw.KEY_DOWN
    LEFT_SHIFT = glfw.KEY_LEFT_SHIFT
    RIGHT_SHIFT = glfw.KEY_RIGHT_SHIFT
    LEFT_CTRL = glfw.KEY_LEFT_CONTROL

    TAB = glfw.KEY_TAB
    COMMA = glfw.KEY_COMMA
    MINUS = glfw.KEY_MINUS
    PERIOD = glfw.KEY_PERIOD
    SLASH = glfw.KEY_SLASH
    SEMICOLON = glfw.KEY_SEMICOLON
    EQUAL = glfw.KEY_EQUAL
    LEFT_BRACKET = glfw.KEY_LEFT_BRACKET
    RIGHT_BRACKET = glfw.KEY_RIGHT_BRACKET
    BACKSLASH = glfw.KEY_BACKSLASH
    BACKSPACE = glfw.KEY_BACKSPACE
    INSERT = glfw.KEY_INSERT
    DELETE = glfw.KEY_DELETE
    HOME = glfw.KEY_HOME
    END = glfw.KEY_END
    CAPS_LOCK = glfw.KEY_CAPS_LOCK

    F1 = glfw.KEY_F1
    F2 = glfw.KEY_F2
    F3 = glfw.KEY_F3
    F4 = glfw.KEY_F4
    F5 = glfw.KEY_F5
    F6 = glfw.KEY_F6
    F7 = glfw.KEY_F7
    F8 = glfw.KEY_F8
    F9 = glfw.KEY_F9
    F10 = glfw.KEY_F10
    F11 = glfw.KEY_F11
    F12 = glfw.KEY_F12

    NUMBER_0 = glfw.KEY_0
    NUMBER_1 = glfw.KEY_1
    NUMBER_2 = glfw.KEY_2
    NUMBER_3 = glfw.KEY_3
    NUMBER_4 = glfw.KEY_4
    NUMBER_5 = glfw.KEY_5
    NUMBER_6 = glfw.KEY_6
    NUMBER_7 = glfw.KEY_7
    NUMBER_8 = glfw.KEY_8
    NUMBER_9 = glfw.KEY_9

    NUMPAD_0 = glfw.KEY_KP_0
    NUMPAD_1 = glfw.KEY_KP_1
    NUMPAD_2 = glfw.KEY_KP_2
    NUMPAD_3 = glfw.KEY_KP_3
    NUMPAD_4 = glfw.KEY_KP_4
    NUMPAD_5 = glfw.KEY_KP_5
    NUMPAD_6 = glfw.KEY_KP_6
    NUMPAD_7 = glfw.KEY_KP_7
    NUMPAD_8 = glfw.KEY_KP_8
    NUMPAD_9 = glfw.KEY_KP_9

    A = glfw.KEY_A
    B = glfw.KEY_B
    C = glfw.KEY_C
    D = glfw.KEY_D
    E = glfw.KEY_E
    F = glfw.KEY_F
    G = glfw.KEY_G
    H = glfw.KEY_H
    I = glfw.KEY_I
    J = glfw.KEY_J
    K = glfw.KEY_K
    L = glfw.KEY_L
    M = glfw.KEY_M
    N = glfw.KEY_N
    O = glfw.KEY_O
    P = glfw.KEY_P
    Q = glfw.KEY_Q
    R = glfw.KEY_R
    S = glfw.KEY_S
    T = glfw.KEY_T
    U = glfw.KEY_U
    V = glfw.KEY_V
    W = glfw.KEY_W
    X = glfw.KEY_X
    Y = glfw.KEY_Y
    Z = glfw.KEY_Z



from pathlib import Path

import moderngl
from headless import WindowConfigTestCase

from moderngl_window import WindowConfig
from moderngl_window.scene import Scene


class WindowConfigTestCase(WindowConfigTestCase):
    class TestConfig(WindowConfig):
        window_size = (16, 32)
        aspect_ratio = 1.0
        gl_version = (4, 1)
        title = "Test"
        resource_dir = Path(__file__).parent / 'fixtures' / 'resources'

    def create_window_config(self, cls):
        """Create a WindowConfig instance passing in the standard params"""
        instance = cls(ctx=self.window.ctx, wnd=self.window, timer=None)
        instance.window_size = self.config.window_size
        instance.aspect_ratio = self.config.aspect_ratio
        instance.gl_version = self.config.gl_version
        return instance

    def test_properties(self):
        """Ensure all callback funcs are callable"""
        # Configured Values
        self.assertIsInstance(self.window.config, WindowConfig)
        self.assertEqual(self.window.size, self.config.window_size)
        self.assertEqual(self.window.width, self.config.window_size[0])
        self.assertEqual(self.window.height, self.config.window_size[1])
        self.assertEqual(self.window.title, self.config.title)
        self.assertEqual(self.window.gl_version, self.config.gl_version)
        self.assertEqual(self.config.aspect_ratio, self.window.aspect_ratio)
        self.assertIsInstance(self.window.ctx, moderngl.Context)
        self.assertIsInstance(self.window.fbo, moderngl.Framebuffer)
        self.assertEqual(self.window.vsync, False)

        # Defaults
        self.assertEqual(self.config.resizable, True)  # Disabled in headless
        self.assertEqual(self.config.cursor, True)  # Disabled in headless
        self.assertEqual(self.config.samples, self.window.samples)
        self.assertIsInstance(self.config.resource_dir, Path)
        self.assertEqual(self.config.clear_color, (0, 0, 0, 0))

        # Ensure callback funcs are actual callable
        self.assertTrue(callable(self.window.resize_func))
        self.assertTrue(callable(self.window.key_event_func))
        self.assertTrue(callable(self.window.mouse_position_event_func))
        self.assertTrue(callable(self.window.mouse_press_event_func))
        self.assertTrue(callable(self.window.mouse_release_event_func))
        self.assertTrue(callable(self.window.mouse_drag_event_func))
        self.assertEqual(self.window.pixel_ratio, 1.0)
        self.assertEqual(self.window.buffer_size, self.config.window_size)
        self.assertEqual(self.window.buffer_width, self.config.window_size[0])
        self.assertEqual(self.window.buffer_height, self.config.window_size[1])

        # Other windows properties
        self.assertEqual(self.window.viewport, (0, 8, 16, 16))
        self.assertEqual(self.window.viewport_size, (16, 16))
        self.assertEqual(self.window.viewport_width, 16)
        self.assertEqual(self.window.viewport_height, 16)
        self.assertEqual(self.window.frames, 0)

        # set properties
        self.window.title = "Modified Title"
        self.assertEqual(self.window.title, "Modified Title")
        value = not self.window.cursor
        self.window.cursor = value
        self.assertEqual(self.window.cursor, value)
        value = self.window.position[0] + 10, self.window.position[1] + 10
        self.window.position = value
        self.assertEqual(self.window.position, value)

    def test_missing_wnd_ctx(self):
        """Attempt creating WindogConfig without a window or ctx"""
        class TestConfig(WindowConfig):
            pass

        with self.assertRaises(ValueError):
            TestConfig(ctx=self.window.ctx)

        with self.assertRaises(ValueError):
            TestConfig(wnd=self.window)

    def test_set_bad_callback(self):
        """Attempt setting bad callbacks"""
        class TextConfig(WindowConfig):
            pass

        with self.assertRaises(ValueError):
            self.window.resize_func = None

        with self.assertRaises(ValueError):
            self.window.resize_func = "Hello"

    def test_load_texture_2d(self):
        """Load texture with shortcut method"""
        texture = self.config.load_texture_2d(
            "textures/crate.png",
            flip=True,
            mipmap_levels=(0, 2),
            anisotropy=4.0,
        )
        self.assertIsInstance(texture, moderngl.Texture)
        self.assertEqual(texture.anisotropy, 4.0)

    def test_load_texture_array(self):
        """Load texture array with shortcut method"""
        texture = self.config.load_texture_array(
            'textures/array.png',
            layers=10,
            flip=True,
            mipmap=False,
            mipmap_levels=(0, 2),
            anisotropy=4.0,
        )
        self.assertIsInstance(texture, moderngl.TextureArray)
        self.assertEqual(texture.anisotropy, 4.0)
        self.assertEqual(texture.layers, 10)

    def test_load_program_single(self):
        """Load a single glsl program"""
        prog = self.config.load_program(path='programs/white.glsl')
        self.assertIsInstance(prog, moderngl.Program)

    def test_load_program_multiple(self):
        """Load program from multiple shader files"""
        prog = self.config.load_program(
            vertex_shader='programs/terrain/terrain_vs.glsl',
            fragment_shader='programs/terrain/terrain_fs.glsl',
            tess_control_shader='programs/terrain/terrain_tc.glsl',
            tess_evaluation_shader='programs/terrain/terrain_te.glsl',
        )
        self.assertIsInstance(prog, moderngl.Program)

    def test_load_text(self):
        """Load text file"""
        text = self.config.load_text('data/data.txt')
        self.assertEqual(text, "Hello")

    def test_load_json(self):
        """Load a json file"""
        json = self.config.load_json('data/data.json')
        self.assertEqual(json, {"test": "Hello"})

    def test_load_binary(self):
        """Load binary file"""
        data = self.config.load_binary('data/data.bin')
        self.assertEqual(data, b'Hello')

    def test_load_scene(self):
        """Load a scene"""
        scene = self.config.load_scene(
            path='scenes/BoxTextured/glTF/BoxTextured.gltf',
            cache=False,
            attr_names=None,
            kind=None,
        )
        self.assertIsInstance(scene, Scene)



"""
Note: In the future we might want to split this into separate scene loaders
"""
from pathlib import Path

from headless import HeadlessTestCase

from moderngl_window import resources
from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.meta import SceneDescription
from moderngl_window.scene import Scene

resources.register_dir((Path(__file__).parent / 'fixtures' / 'resources').resolve())


class SceneLoadersTestCase(HeadlessTestCase):
    window_size = (16, 16)
    aspect_ratio = 1.0

    def test_wavefront(self):
        """Load wavefront file"""
        scene = resources.scenes.load(SceneDescription(path='scenes/crate/crate.obj'))
        self.assertIsInstance(scene, Scene)

    def test_wavefont_not_found(self):
        """Ensure ImproperlyConfigured is raised when wavefront is not found"""
        with self.assertRaises(ImproperlyConfigured):
            resources.scenes.load(SceneDescription(path='scenes/doesnotexist.obj'))

    def test_gltf(self):
        """Load standard gltf"""
        scene = resources.scenes.load(SceneDescription(path='scenes/BoxTextured/glTF/BoxTextured.gltf'))
        self.assertIsInstance(scene, Scene)

    def test_gltf_binary(self):
        """Load binary gltf"""
        scene = resources.scenes.load(SceneDescription(path='scenes/BoxTextured/glTF-Binary/BoxTextured.glb'))
        self.assertIsInstance(scene, Scene)

    def test_gltf_embedded(self):
        """Load embedded gltf"""
        scene = resources.scenes.load(SceneDescription(path='scenes/BoxTextured/glTF-Embedded/BoxTextured.gltf'))
        self.assertIsInstance(scene, Scene)

    def test_gltf_not_found(self):
        """Attempt to load nonexisting gltf"""
        with self.assertRaises(ImproperlyConfigured):
            resources.scenes.load(SceneDescription(path='scenes/doesnotexist.gltf'))

    def test_stl(self):
        scene = resources.scenes.load(SceneDescription(path='scenes/uplink.stl'))
        self.assertIsInstance(scene, Scene)



import moderngl_window as mglw
from moderngl_window.scene.camera import KeyboardCamera


class CameraWindow(mglw.WindowConfig):
    """Base class with built in 3D camera support"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.camera = KeyboardCamera(self.wnd.keys, aspect_ratio=self.wnd.aspect_ratio)
        self.camera_enabled = True

    def on_key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        if self.camera_enabled:
            self.camera.key_input(key, action, modifiers)

        if action == keys.ACTION_PRESS:
            if key == keys.C:
                self.camera_enabled = not self.camera_enabled
                self.wnd.mouse_exclusivity = self.camera_enabled
                self.wnd.cursor = not self.camera_enabled
            if key == keys.SPACE:
                self.timer.toggle_pause()

    def on_mouse_position_event(self, x: int, y: int, dx, dy):
        if self.camera_enabled:
            self.camera.rot_state(-dx, -dy)

    def on_resize(self, width: int, height: int):
        self.camera.projection.update(aspect_ratio=self.wnd.aspect_ratio)



import sched
import time
from typing import Any, Callable

from moderngl_window.timers.base import BaseTimer


class Scheduler:
    def __init__(self, timer: BaseTimer):
        """Create a Scheduler object to handle events.

        Args:
            timer (BaseTimer): timer to use, subclass of BaseTimer.

        Raises:
            ValueError: timer is not a valid argument.
        """
        if not isinstance(timer, BaseTimer):
            raise ValueError(
                "timer, {}, has to be a instance of BaseTimer or a callable!".format(timer)
            )

        self._events: dict[int, sched.Event] = dict()
        self._event_id = 0

        self._scheduler = sched.scheduler(lambda: timer.time, time.sleep)

    def run_once(
        self,
        action: Callable[[Any], Any],
        delay: float,
        *,
        priority: int = 1,
        arguments: tuple[Any, ...] = (),
        kwargs: dict[Any, Any] = dict(),
    ) -> int:
        """Schedule a function for execution after a delay.

        Args:
            action (callable):
                function to be called.
            delay (float):
                delay in seconds.
            priority (int, optional):
                priority for this event, lower is more important. Defaults to 1.
            arguments (tuple, optional):
                arguments for the action. Defaults to ().
            kwargs (dict, optional):
                keyword arguments for the action. Defaults to dict().

        Returns:
            int: event id that can be canceled.
        """
        event = self._scheduler.enter(delay, priority, action, arguments, kwargs)
        self._events[self._event_id] = event
        self._event_id += 1
        return self._event_id - 1

    def run_at(
        self,
        action: Callable[[Any], Any],
        time: float,
        *,
        priority: int = 1,
        arguments: tuple[Any, ...] = (),
        kwargs: dict[Any, Any] = dict(),
    ) -> int:
        """Schedule a function to be executed at a certain time.

        Args:
            action (callable):
                function to be called.
            time (float):
                epoch time at which the function should be called.
            priority (int, optional):
                priority for this event, lower is more important. Defaults to 1.
            arguments (tuple, optional):
                arguments for the action. Defaults to ().
            kwargs (dict, optional):
                keyword arguments for the action. Defaults to dict().

        Returns:
            int: event id that can be canceled.
        """
        event = self._scheduler.enterabs(time, priority, action, arguments, kwargs)
        self._events[self._event_id] = event
        self._event_id += 1
        return self._event_id - 1

    def run_every(
        self,
        action: Callable[[Any], Any],
        delay: float,
        *,
        priority: int = 1,
        initial_delay: float = 0.0,
        arguments: tuple[Any, ...] = (),
        kwargs: dict[Any, Any] = dict(),
    ) -> int:
        """Schedule a recurring function to be called every `delay` seconds after a initial delay.

        Args:
            action (callable):
                function to be called.
            delay (float):
                delay in seconds.
            priority (int, optional):
                priority for this event, lower is more important. Defaults to 1.
            initial_delay (float, optional):
                initial delay in seconds before executing for the first time.
            Defaults to 0. arguments (tuple, optional):
                arguments for the action. Defaults to ().
            kwargs (dict, optional):
                keyword arguments for the action. Defaults to dict().

        Returns:
            int: event id that can be canceled.
        """
        recurring_event = self._recurring_event_factory(
            action, arguments, kwargs, (delay, priority), self._event_id
        )
        event = self._scheduler.enter(initial_delay, priority, recurring_event)
        self._events[self._event_id] = event
        self._event_id += 1
        return self._event_id - 1

    def _recurring_event_factory(
        self,
        function: Callable[[Any], Any],
        arguments: tuple[Any, ...],
        kwargs: dict[Any, Any],
        scheduling_info: tuple[Any, Any],
        id: int,
    ) -> Callable[[], None]:
        """Factory for creating recurring events that will reschedule themselves.

        Args:
            function (callable): function to be called.
            arguments (tuple): arguments for the function.
            kwargs (dict): keyword arguments for the function.
            scheduling_info (tuple): tuple of information for scheduling the task.
            id (int): event id this event should be assigned to.
        """

        def _f() -> None:
            function(*arguments, **kwargs)
            event = self._scheduler.enter(*scheduling_info, _f)
            self._events[id] = event

        return _f

    def execute(self) -> None:
        """Run the scheduler without blocking and execute any expired events."""
        self._scheduler.run(blocking=False)

    def cancel(self, event_id: int, delay: float = 0) -> None:
        """Cancel a previously scheduled event.

        Args:
            event_id (int): event to be canceled
            delay (float, optional): delay before canceling the event. Defaults to 0.
        """
        if delay == 0:
            self._cancel(event_id)
        else:
            self.run_once(self._cancel, delay, priority=0, arguments=(event_id,))

    def _cancel(self, event_id: int) -> None:
        if event_id not in self._events:
            raise ValueError("Recurring event with id {} does not exist".format(event_id))
        event = self._events.pop(event_id)
        self._scheduler.cancel(event)



from __future__ import annotations

import os
from typing import Any, Optional

import glm
import moderngl

import moderngl_window
from moderngl_window.conf import settings
from moderngl_window.meta import ProgramDescription
from moderngl_window.resources.programs import programs

from .mesh import Mesh

settings.PROGRAM_DIRS.append(os.path.join(os.path.dirname(__file__), "programs"))


class MeshProgram:
    """
    Describes how a mesh is rendered using a specific shader program
    """

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        """Initialize.

        Args:
            program: The moderngl program
        """
        self.program = program

    @property
    def ctx(self) -> moderngl.Context:
        """moderngl.Context: The current context"""
        return moderngl_window.ctx()

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        """Draw code for the mesh

        Args:
            mesh (Mesh): The mesh to render
        Keyword Args:
            projection_matrix (numpy.ndarray): projection_matrix (bytes)
            model_matrix (numpy.ndarray): view_matrix (bytes)
            camera_matrix (numpy.ndarray): camera_matrix (bytes)
            time (float): The current time
        """
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"
        self.program["m_proj"].write(projection_matrix)
        self.program["m_mv"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> MeshProgram | None:
        """
        Determine if this ``MeshProgram`` should be applied to the mesh.
        Can return self or some ``MeshProgram`` instance to support dynamic ``MeshProgram`` creation

        Args:
            mesh: The mesh to inspect
        """
        raise NotImplementedError(
            "apply is not implemented. Please override the MeshProgram method"
        )


class VertexColorProgram(MeshProgram):
    """Vertex color program"""

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        super().__init__(program=None)
        self.program = programs.load(ProgramDescription(path="scene_default/vertex_color.glsl"))

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> Optional[MeshProgram]:
        if not mesh.material:
            return None

        if mesh.attributes.get("TEXCOORD_0"):
            return None

        if mesh.attributes.get("COLOR_0"):
            return self

        return None


class ColorLightProgram(MeshProgram):
    """Simple color program with light"""

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        super().__init__(program=None)
        self.program = programs.load(ProgramDescription(path="scene_default/color_light.glsl"))

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"
        if mesh.material is not None:
            # if mesh.material.double_sided:
            #     self.ctx.disable(moderngl.CULL_FACE)
            # else:
            #     self.ctx.enable(moderngl.CULL_FACE)

            if mesh.material.color:
                self.program["color"].value = tuple(mesh.material.color)
            else:
                self.program["color"].value = (1.0, 1.0, 1.0, 1.0)

        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> MeshProgram | None:
        if not mesh.material:
            return None

        if not mesh.attributes.get("NORMAL"):
            return None

        return self


class TextureProgram(MeshProgram):
    """Plan textured"""

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        super().__init__(program=None)
        self.program = programs.load(ProgramDescription(path="scene_default/texture.glsl"))

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"
        assert mesh.material is not None, "There is no material to render"
        assert (
            mesh.material.mat_texture is not None
        ), "The material does not have a texture to render"
        assert (
            mesh.material.mat_texture.texture is not None
        ), "The material texture is not linked to a texture, so it can not be rendered"

        mesh.material.mat_texture.texture.use()
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> Optional[MeshProgram]:
        if not mesh.material:
            return None

        if mesh.attributes.get("NORMAL"):
            return None

        if not mesh.attributes.get("TEXCOORD_0"):
            return None

        if mesh.attributes.get("COLOR_0"):
            return None

        if mesh.material.mat_texture is not None:
            return self

        return None


class TextureVertexColorProgram(MeshProgram):
    """textured object with vertex color"""

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/vertex_color_texture.glsl")
        )

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"
        assert mesh.material is not None, "There is no material to render"
        assert (
            mesh.material.mat_texture is not None
        ), "The material does not have a texture to render"
        assert (
            mesh.material.mat_texture.texture is not None
        ), "The material texture is not linked to a texture, so it can not be rendered"

        mesh.material.mat_texture.texture.use()
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> MeshProgram | None:
        if not mesh.material:
            return None

        if mesh.attributes.get("NORMAL"):
            return None

        if not mesh.attributes.get("TEXCOORD_0"):
            return None

        if not mesh.attributes.get("COLOR_0"):
            return None

        if mesh.material.mat_texture is not None:
            return self

        return None


class TextureLightProgram(MeshProgram):
    """
    Simple texture program
    """

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        super().__init__(program=None)
        self.program = programs.load(ProgramDescription(path="scene_default/texture_light.glsl"))

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"
        assert mesh.material is not None, "There is no material to render"
        assert (
            mesh.material.mat_texture is not None
        ), "The material does not have a texture to render"
        assert (
            mesh.material.mat_texture.texture is not None
        ), "The material texture is not linked to a texture, so it can not be rendered"

        # if mesh.material.double_sided:
        #     self.ctx.disable(moderngl.CULL_FACE)
        # else:
        #     self.ctx.enable(moderngl.CULL_FACE)

        mesh.material.mat_texture.texture.use()
        self.program["texture0"].value = 0
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> MeshProgram | None:
        if not mesh.material:
            return None

        if not mesh.attributes.get("NORMAL"):
            return None

        if not mesh.attributes.get("TEXCOORD_0"):
            return None

        if mesh.material.mat_texture is not None:
            return self

        return None


class TextureLightColorProgram:
    pass


class FallbackProgram(MeshProgram):
    """
    Fallback program only rendering positions in white
    """

    def __init__(self, program: Optional[moderngl.Program] = None, **kwargs: Any) -> None:
        super().__init__(program=None)
        self.program = programs.load(ProgramDescription(path="scene_default/fallback.glsl"))

    def draw(
        self,
        mesh: Mesh,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        assert self.program is not None, "There is no program to draw"
        assert mesh.vao is not None, "There is no vao to render"

        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)

        if mesh.material:
            self.program["color"].value = tuple(mesh.material.color[0:3])
        else:
            self.program["color"].value = (1.0, 1.0, 1.0)

        mesh.vao.render(self.program)

    def apply(self, mesh: Mesh) -> MeshProgram | None:
        return self



from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Union

from moderngl_window.conf import settings


@contextmanager
def texture_dirs(paths: list[Union[Path, str]]) -> Generator[None, None, None]:
    """Context manager temporarily replacing texture paths
    Args:
        paths (list[Union[Path, str]]): list of paths
    """
    original_dirs = settings.DATA_DIRS
    settings.TEXTURE_DIRS = paths
    yield None
    settings.TEXTURE_DIRS = original_dirs



from pathlib import Path

from moderngl_window.finders import texture
from moderngl_window.loaders.base import BaseLoader
from moderngl_window.meta.base import ResourceDescription
from moderngl_window.meta.texture import TextureDescription


class IconLoader(BaseLoader):
    kind = "icon"
    meta: TextureDescription

    def __init__(self, meta: ResourceDescription) -> None:
        super().__init__(meta)

    def find_icon(self) -> Path:
        """Find resource using texture finders.

        This is mainly a shortcut method to simplify the task.

        Args:
            path: Path to resource
        """
        abs_path = self._find(self.meta.path, texture.get_finders())
        if abs_path is None:
            raise ValueError("Could not find the icon specified. {}".format(self.meta.path))
        return abs_path



from typing import Any, Optional

from moderngl_window.meta.base import ResourceDescription


class DataDescription(ResourceDescription):
    """Describes data file to load.

    This is a generic resource description type
    for loading resources that are not textures, programs and scenes.
    That loaded class is used depends on the ``kind`` or the file extension.

    Currently used to load:

    - text files
    - json files
    - binary files

    .. code:: python

        # Describe a text file. Text loader is used based on file extension
        DataDescription(path='data/text.txt')

        # Describe a json file. Json loader is used based on file extension
        DataDescription(path='data/data.json')

        # Describe a binary file. Specify a binary loader should be used.
        DataDescription(path='data/data.bin', kind='binary')
    """

    default_kind: str = ""
    resource_type = "data"

    def __init__(
        self, path: Optional[str] = None, kind: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Initialize the resource description.

        Keyword Args:
            path (str): Relative path to the resource
            kind (str): The resource kind deciding loader class
            **kwargs: Additional custom attributes
        """
        kwargs.update({"path": path, "kind": kind})
        super().__init__(**kwargs)



from .camera import Camera as Camera
from .camera import KeyboardCamera as KeyboardCamera
from .camera import OrbitCamera as OrbitCamera
from .material import Material as Material
from .material import MaterialTexture as MaterialTexture
from .mesh import Mesh as Mesh
from .node import Node as Node
from .programs import MeshProgram as MeshProgram
from .scene import Scene as Scene

__all__ = [
    "Camera",
    "KeyboardCamera",
    "OrbitCamera",
    "Material",
    "MaterialTexture",
    "Mesh",
    "Node",
    "MeshProgram",
    "Scene",
]



from pathlib import Path

import glm
import moderngl
import numpy
from base import CameraWindow

import moderngl_window


class LinesDemo(CameraWindow):
    """Rendering thick lines with geometry shader

    Example is basic and incomplete, but shows how
    one could use the geometry shader to create simple thick lines.
    """

    gl_version = (3, 3)
    title = "Thick Lines"
    resource_dir = (Path(__file__) / "../resources").absolute()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True

        self.prog = self.load_program("programs/lines/lines.glsl")
        self.prog["color"].value = (1.0, 1.0, 1.0, 1.0)
        self.prog["m_model"].write(glm.translate(glm.vec3(0.0, 0.0, -3.5)))

        N = 10

        # Create lines geometry
        def gen_lines():
            for i in range(N):
                # A
                yield -1.0
                yield 1.0 - i * 2.0 / N
                yield 0.0
                # b
                yield 1.0
                yield 1.0 - i * 2.0 / N
                yield 0.0

        buffer = self.ctx.buffer(numpy.fromiter(gen_lines(), dtype="f4", count=N * 6).tobytes())
        self.lines = self.ctx.vertex_array(
            self.prog,
            [
                (buffer, "3f", "in_position"),
            ],
        )

    def on_render(self, time: float, frametime: float):
        # self.ctx.enable_only(moderngl.DEPTH_TEST)

        self.prog["m_proj"].write(self.camera.projection.matrix)
        self.prog["m_cam"].write(self.camera.matrix)
        self.lines.render(mode=moderngl.LINES)


if __name__ == "__main__":
    moderngl_window.run_window_config(LinesDemo)



from pathlib import Path
from unittest import TestCase

from moderngl_window import resources
from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.meta import DataDescription

resources.register_dir((Path(__file__).parent / 'fixtures' / 'resources').resolve())


class DataLoaderTestcase(TestCase):

    def test_txt(self):
        """Ensure correct loader is selected by looking at file extension (txt)"""
        text = resources.data.load(DataDescription(path='data/data.txt'))
        self.assertEqual(text, "Hello")

    def test_json(self):
        """Ensure correct loader is selected by looking at file extension (json)"""
        json = resources.data.load(DataDescription(path='data/data.json'))
        self.assertEqual(json, {"test": "Hello"})

    def test_binary_kind(self):
        """Loading a binary file"""
        data = resources.data.load(DataDescription(path='data/data.bin', kind="binary"))
        self.assertEqual(data, b'Hello')

    def test_text_kind(self):
        """Load a e textfile"""
        text = resources.data.load(DataDescription(path='data/data.txt', kind="text"))
        self.assertEqual(text, "Hello")

    def test_json_kind(self):
        """Load a json file"""
        json = resources.data.load(DataDescription(path='data/data.json', kind="json"))
        self.assertEqual(json, {"test": "Hello"})

    def test_txt_not_found(self):
        """Ensure ImproperlyConfigured is raised if file is not found"""
        with self.assertRaises(ImproperlyConfigured):
            resources.data.load(DataDescription(path='data/notfound.txt'))

    def test_json_not_found(self):
        """Ensure ImproperlyConfigured is raised if file is not found"""
        with self.assertRaises(ImproperlyConfigured):
            resources.data.load(DataDescription(path='data/notfound.json'))

    def test_binary_not_found(self):
        """Ensure ImproperlyConfigured is raised if file is not found"""
        with self.assertRaises(ImproperlyConfigured):
            resources.data.load(DataDescription(path='data/notfound.bin', kind="binary"))

    def test_binary_abspath(self):
        """Strip search directories and use absolute path"""
        path = (Path(__file__).parent / "fixtures/resources/data/data.json").resolve()
        with resources.temporary_dirs([]):
            json = resources.data.load(DataDescription(path=path, kind="json"))
            self.assertEqual(json, {"test": "Hello"})



import logging
from pathlib import Path
from typing import Optional, Union

try:
    from PIL import Image
except ImportError as ex:
    raise ImportError("Texture loader 'PillowLoader' requires Pillow: {}".format(ex))

from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.loaders.base import BaseLoader
from moderngl_window.meta.base import ResourceDescription
from moderngl_window.meta.texture import TextureDescription
from moderngl_window.resources.textures import TextureAny

logger = logging.getLogger(__name__)


class PillowLoader(BaseLoader):
    """Base loader using PIL/Pillow"""

    kind = "__unknown__"
    image: Image.Image
    meta: TextureDescription

    def __init__(self, meta: ResourceDescription):
        super().__init__(meta)

    def load(self) -> TextureAny:
        raise NotImplementedError()

    def _open_image(self) -> Image.Image:
        if self.meta.image:
            self.image = self.meta.image
        else:
            self.meta.resolved_path = self.find_texture(self.meta.path)
            logger.info("loading %s", self.meta.resolved_path)
            if not self.meta.resolved_path:
                raise ImproperlyConfigured("Cannot find texture: {}".format(self.meta.path))

            self.image = Image.open(self.meta.resolved_path)

            # If the image is animated (like a gif anim) we convert it into a vertical strip
            if (
                hasattr(self.image, "is_animated")
                and self.image.is_animated
                and hasattr(self.image, "n_frames")
            ):
                self.layers = self.image.n_frames
                anim = Image.new(
                    self.image.palette.mode if self.image.palette is not None else "L",
                    (self.image.width, self.image.height * self.image.n_frames),
                )
                anim.putalpha(0)

                for frame_number in range(self.image.n_frames):
                    self.image.seek(frame_number)
                    frame = self._palette_to_raw(self.image, mode="RGBA")
                    anim.paste(frame, (0, frame_number * self.image.height))

                self.image = anim

        self.image = self._apply_modifiers(self.image)
        return self.image

    def _load_texture(self, path: Union[str, Path]) -> Image.Image:
        """Find and load separate texture. Useful when multiple textue files needs to be loaded"""
        resolved_path = self.find_texture(path)
        logger.info("loading %s", resolved_path)
        if not resolved_path:
            raise ImproperlyConfigured("Cannot find texture: {}".format(path))

        image = Image.open(resolved_path)
        return self._apply_modifiers(image)

    def _apply_modifiers(self, image: Image.Image) -> Image.Image:
        if self.meta.flip_x:
            image = image.transpose(Image.Transpose.FLIP_LEFT_RIGHT)

        if self.meta.flip_y:
            image = image.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

        return self._palette_to_raw(image)

    def _palette_to_raw(self, image: Image.Image, mode: Optional[str] = None) -> Image.Image:
        """Converts image to raw if palette is present"""
        if image.palette and image.palette.mode.lower() in ["rgb", "rgba"]:
            mode = mode or image.palette.mode
            logger.debug("Converting P image to %s using palette", mode)
            return image.convert(mode)

        return image

    def _close_image(self) -> None:
        self.image.close()


def image_data(image: Image.Image) -> tuple[int, bytes]:
    """Get components and bytes for an image.
    The number of components is assumed by image
    size and the byte length of the raw data.

    Returns:
        tuple[int, bytes]: Number of components, byte data
    """
    # NOTE: We might want to check the actual image.mode
    #       and convert to an acceptable format.
    #       At the moment we load the data as is.
    data = image.tobytes()
    components = len(data) // (image.size[0] * image.size[1])
    logger.debug(
        "image_data size=[%s, %s] components=%s bytes=%s",
        image.size[0],
        image.size[1],
        components,
        len(data),
    )
    return components, data



from pathlib import Path
from unittest import TestCase

from utils import settings_context

from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.finders import data, program, scene, texture
from moderngl_window.finders.base import BaseFilesystemFinder


class FinderTestCase(TestCase):
    root = Path(__file__).parent / 'fixtures' / 'resources'
    finder_settings = {
        'DATA_DIRS': [Path(root, 'data')],
        'PROGRAM_DIRS': [Path(root, 'programs')],
        'TEXTURE_DIRS': [Path(root, 'textures')],
        'SCENE_DIRS': [Path(root, 'scenes')],
    }

    def test_data_finder(self):
        """Find a data file"""
        with settings_context(self.finder_settings):
            result = data.FilesystemFinder().find(Path('data.json'))
            self.assertIsInstance(result, Path)
            self.assertTrue(result.name, 'data.json')

    def test_program_finder(self):
        """Find a glsl file"""
        with settings_context(self.finder_settings):
            result = program.FilesystemFinder().find(Path('test.glsl'))
            self.assertIsInstance(result, Path)
            self.assertTrue(result.name, 'test.glsl')

    def test_texture_finder(self):
        """Find a texture"""
        with settings_context(self.finder_settings):
            result = texture.FilesystemFinder().find(Path('image.png'))
            self.assertIsInstance(result, Path)
            self.assertTrue(result.name, 'image.png')

    def test_scene_finder(self):
        """Find a scene"""
        with settings_context(self.finder_settings):
            result = scene.FilesystemFinder().find(Path('model.obj'))
            self.assertIsInstance(result, Path)
            self.assertTrue(result.name, 'model.obj')

    def test_relative_path_raises_exception(self):
        with settings_context({'DATA_DIRS': ['relative_location']}):
            with self.assertRaises(ImproperlyConfigured):
                data.FilesystemFinder().find(Path('something'))

    def test_absolute_path(self):
        """Ensure absolute paths are ignored"""
        with settings_context(self.finder_settings):
            finder = data.FilesystemFinder()
            result = finder.find(Path(self.root, Path('data/data.json')))
            self.assertIsNone(result)

    def test_not_found(self):
        """Ensure finder returns non when nothing was found"""
        finder = data.FilesystemFinder()
        result = finder.find(Path(self.root, Path('data/idontexist.json')))
        self.assertIsNone(result)

    def test_no_search_dirs(self):
        """When no search dirs the finder should return None"""
        with settings_context({'DATA_DIRS': []}):
            finder = data.FilesystemFinder()
            result = finder.find(Path('data/data.json'))
            self.assertIsNone(result)

    def test_non_path(self):
        """Raise ValueError if finder gets non-Path instance"""
        finder = data.FilesystemFinder()
        with self.assertRaises(ValueError):
            finder.find('test')

    def test_custom_finder_missing_setting(self):
        """Ensure broken finder with nonexisting settings attr is detected"""
        class BrokenFinder(BaseFilesystemFinder):
            settings_attr = "NOPE"

        with self.assertRaises(ImproperlyConfigured):
            BrokenFinder()



import logging

import moderngl

from moderngl_window.loaders.texture.pillow import PillowLoader, image_data

logger = logging.getLogger(__name__)


class Loader(PillowLoader):
    kind = "2d"

    def load(self) -> moderngl.Texture:
        """Load a 2d texture as configured in the supplied ``TextureDescription``

        Returns:
            moderngl.Texture: The Texture instance
        """
        self._open_image()

        components, data = image_data(self.image)

        texture = self.ctx.texture(
            self.image.size,
            components,
            data,
        )
        texture.extra = {"meta": self.meta}

        if self.meta.mipmap_levels is not None:
            self.meta.mipmap = True

        if self.meta.mipmap:
            if isinstance(self.meta.mipmap_levels, tuple):
                texture.build_mipmaps(*self.meta.mipmap_levels)
            else:
                texture.build_mipmaps()

            if self.meta.anisotropy:
                texture.anisotropy = self.meta.anisotropy

        self._close_image()

        return texture



from .base import ResourceDescription as ResourceDescription
from .data import DataDescription as DataDescription
from .program import ProgramDescription as ProgramDescription
from .scene import SceneDescription as SceneDescription
from .texture import TextureDescription as TextureDescription



from typing import Any


class KeyModifiers:
    """Namespace for storing key modifiers"""

    shift: Any = False
    ctrl: Any = False
    alt: Any = False

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return "<KeyModifiers shift={} ctrl={} alt={}>".format(self.shift, self.ctrl, self.alt)


class BaseKeys:
    """
    Namespace for mapping key constants.
    This is simply a template for what keys should be mapped for all window libraries
    """

    # Fallback press/release action when window libraries don't have this
    ACTION_PRESS: Any = "ACTION_PRESS"
    ACTION_RELEASE: Any = "ACTION_RELEASE"

    ESCAPE: Any = "undefined"
    SPACE: Any = "undefined"
    ENTER: Any = "undefined"
    PAGE_UP: Any = "undefined"
    PAGE_DOWN: Any = "undefined"
    LEFT: Any = "undefined"
    RIGHT: Any = "undefined"
    UP: Any = "undefined"
    DOWN: Any = "undefined"
    LEFT_SHIFT: Any = "undefined"
    RIGHT_SHIFT: Any = "undefined"
    LEFT_CTRL: Any = "undefined"

    TAB: Any = "undefined"
    COMMA: Any = "undefined"
    MINUS: Any = "undefined"
    PERIOD: Any = "undefined"
    SLASH: Any = "undefined"
    SEMICOLON: Any = "undefined"
    EQUAL: Any = "undefined"
    LEFT_BRACKET: Any = "undefined"
    RIGHT_BRACKET: Any = "undefined"
    BACKSLASH: Any = "undefined"
    BACKSPACE: Any = "undefined"
    INSERT: Any = "undefined"
    DELETE: Any = "undefined"
    HOME: Any = "undefined"
    END: Any = "undefined"
    CAPS_LOCK: Any = "undefined"

    F1: Any = "undefined"
    F2: Any = "undefined"
    F3: Any = "undefined"
    F4: Any = "undefined"
    F5: Any = "undefined"
    F6: Any = "undefined"
    F7: Any = "undefined"
    F8: Any = "undefined"
    F9: Any = "undefined"
    F10: Any = "undefined"
    F11: Any = "undefined"
    F12: Any = "undefined"

    NUMBER_0: Any = "undefined"
    NUMBER_1: Any = "undefined"
    NUMBER_2: Any = "undefined"
    NUMBER_3: Any = "undefined"
    NUMBER_4: Any = "undefined"
    NUMBER_5: Any = "undefined"
    NUMBER_6: Any = "undefined"
    NUMBER_7: Any = "undefined"
    NUMBER_8: Any = "undefined"
    NUMBER_9: Any = "undefined"

    NUMPAD_0: Any = "undefined"
    NUMPAD_1: Any = "undefined"
    NUMPAD_2: Any = "undefined"
    NUMPAD_3: Any = "undefined"
    NUMPAD_4: Any = "undefined"
    NUMPAD_5: Any = "undefined"
    NUMPAD_6: Any = "undefined"
    NUMPAD_7: Any = "undefined"
    NUMPAD_8: Any = "undefined"
    NUMPAD_9: Any = "undefined"

    A: Any = "undefined"
    B: Any = "undefined"
    C: Any = "undefined"
    D: Any = "undefined"
    E: Any = "undefined"
    F: Any = "undefined"
    G: Any = "undefined"
    H: Any = "undefined"
    I: Any = "undefined"
    J: Any = "undefined"
    K: Any = "undefined"
    L: Any = "undefined"
    M: Any = "undefined"
    N: Any = "undefined"
    O: Any = "undefined"
    P: Any = "undefined"
    Q: Any = "undefined"
    R: Any = "undefined"
    S: Any = "undefined"
    T: Any = "undefined"
    U: Any = "undefined"
    V: Any = "undefined"
    W: Any = "undefined"
    X: Any = "undefined"
    Y: Any = "undefined"
    Z: Any = "undefined"



from typing import Optional

import moderngl
import numpy

from moderngl_window.geometry.attributes import AttributeNames
from moderngl_window.opengl.vao import VAO


def quad_fs(
    attr_names: type[AttributeNames] = AttributeNames,
    normals: bool = True,
    uvs: bool = True,
    name: Optional[str] = None,
) -> VAO:
    """
    Creates a screen aligned quad using two triangles with normals and texture coordinates.

    Keyword Args:
        attr_names (AttributeNames): Attrib name config
        normals (bool): Include normals in VAO
        uvs (bool): Include texture coordinates in VAO
        name (str): Optional name for the VAO
    Returns:
        A :py:class:`~moderngl_window.opengl.vao.VAO` instance.
    """
    return quad_2d(
        size=(2.0, 2.0),
        normals=normals,
        uvs=uvs,
        attr_names=attr_names,
        name=name,
    )


def quad_2d(
    size: tuple[float, float] = (1.0, 1.0),
    pos: tuple[float, float] = (0.0, 0.0),
    normals: bool = True,
    uvs: bool = True,
    attr_names: type[AttributeNames] = AttributeNames,
    name: Optional[str] = None,
) -> VAO:
    """
    Creates a 2D quad VAO using 2 triangles with normals and texture coordinates.

    Keyword Args:
        size (tuple): width and height
        pos (float): Center position x and y
        normals (bool): Include normals in VAO
        uvs (bool): Include texture coordinates in VAO
        attr_names (AttributeNames): Attrib name config
        name (str): Optional name for the VAO
    Returns:
        A :py:class:`~moderngl_window.opengl.vao.VAO` instance.
    """
    width, height = size
    xpos, ypos = pos

    # fmt: off
    pos_data = numpy.array([
        xpos - width / 2.0, ypos + height / 2.0, 0.0,
        xpos - width / 2.0, ypos - height / 2.0, 0.0,
        xpos + width / 2.0, ypos - height / 2.0, 0.0,
        xpos - width / 2.0, ypos + height / 2.0, 0.0,
        xpos + width / 2.0, ypos - height / 2.0, 0.0,
        xpos + width / 2.0, ypos + height / 2.0, 0.0,
    ], dtype=numpy.float32)

    normal_data = numpy.array([
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
    ], dtype=numpy.float32)

    uv_data = numpy.array([
        0.0, 1.0,
        0.0, 0.0,
        1.0, 0.0,
        0.0, 1.0,
        1.0, 0.0,
        1.0, 1.0,
    ], dtype=numpy.float32)
    # fmt: on

    vao = VAO(name or "geometry:quad", mode=moderngl.TRIANGLES)
    vao.buffer(pos_data, "3f", [attr_names.POSITION])
    if normals:
        vao.buffer(normal_data, "3f", [attr_names.NORMAL])
    if uvs:
        vao.buffer(uv_data, "2f", [attr_names.TEXCOORD_0])

    return vao



# Window / context properties
WINDOW = {
    "gl_version": (3, 3),
    "class": "moderngl_window.context.pyglet.Window",
    "size": (1280, 720),
    "aspect_ratio": 16 / 9,
    "fullscreen": False,
    "resizable": True,
    "title": "ModernGL Window",
    "vsync": True,
    "cursor": True,
    "samples": 0,
}

SCREENSHOT_PATH = None

# Finders
PROGRAM_FINDERS = [
    "moderngl_window.finders.program.FilesystemFinder",
]

TEXTURE_FINDERS = [
    "moderngl_window.finders.texture.FilesystemFinder",
]

SCENE_FINDERS = [
    "moderngl_window.finders.scene.FilesystemFinder",
]

DATA_FINDERS = [
    "moderngl_window.finders.data.FilesystemFinder",
]

# Finder directories: Where finders look for their resources
PROGRAM_DIRS: list[str] = []
TEXTURE_DIRS: list[str] = []
SCENE_DIRS: list[str] = []
DATA_DIRS: list[str] = []


# Loaders
PROGRAM_LOADERS = [
    "moderngl_window.loaders.program.single.Loader",
    "moderngl_window.loaders.program.separate.Loader",
]

TEXTURE_LOADERS = [
    "moderngl_window.loaders.texture.t2d.Loader",
    "moderngl_window.loaders.texture.array.Loader",
    "moderngl_window.loaders.texture.cube.Loader",
]

SCENE_LOADERS = [
    "moderngl_window.loaders.scene.gltf2.Loader",
    "moderngl_window.loaders.scene.wavefront.Loader",
    "moderngl_window.loaders.scene.stl.Loader",
]

DATA_LOADERS = [
    "moderngl_window.loaders.data.binary.Loader",
    "moderngl_window.loaders.data.text.Loader",
    "moderngl_window.loaders.data.json.Loader",
]



import moderngl
import numpy
from headless import HeadlessTestCase

from moderngl_window.opengl.vao import VAO, BufferInfo, VAOError


class VaoTestCase(HeadlessTestCase):

    def createProgram(self):
        return self.ctx.program(
            vertex_shader="""
            #version 330

            in vec3 position;
            in vec3 normal;
            out vec3 v_normal;

            void main() {
                gl_Position = vec4(position, 1.0);
                v_normal = normalize(normal);
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;
            in vec3 v_normal;

            void main() {
                float col = length(v_normal);
                fragColor = vec4(col);
            }
            """,
        )

    def test_create(self):
        mesh = VAO("test", mode=moderngl.LINES)
        mesh.buffer(
            numpy.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0], dtype="f4"), "3f", "position"
        )
        mesh.buffer(
            numpy.array([0.0, 0.0, 1.0, 1.0, 0.0, 1.0], dtype="f4"), "3f", "normal"
        )
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0], dtype="f4"), "2f", "uv")

        # Ensure basic properties are correct
        self.assertEqual(mesh.name, "test")
        self.assertEqual(mesh.vertex_count, 2)
        self.assertEqual(mesh.mode, moderngl.LINES)

        # Ensure buffers are present
        self.assertIsInstance(mesh.get_buffer_by_name("position"), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name("normal"), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name("uv"), BufferInfo)
        self.assertEqual(mesh.get_buffer_by_name("something"), None)

        # Create a progam using a subset of the buffers
        prog = self.createProgram()
        vao = mesh.instance(prog)
        self.assertIsInstance(vao, moderngl.VertexArray)

        # Render directly with VAO and VertexArray instance
        mesh.render(prog)
        vao.render()

    def test_illegal_draw_mode(self):
        """Create vao with illegal draw mode"""
        with self.assertRaises(VAOError):
            VAO(mode=1337)

    def test_add_illegal_buffer(self):
        """Attempt to add illegal buffer"""
        vao = VAO()
        with self.assertRaises(VAOError):
            vao.buffer("stuff", "1f", "in_position")

    def test_normalized_types(self):
        """Ensure VAO wrapper can handle normalized types"""
        # prog = self.createProgram()
        mesh = VAO("test", mode=moderngl.LINES)
        mesh.buffer(
            numpy.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0], dtype="i4"), "3ni", "position"
        )
        mesh.buffer(
            numpy.array([0.0, 0.0, 1.0, 1.0, 0.0, 1.0], dtype="f4"), "3nf", "normal"
        )
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0], dtype="f4"), "2f", "uv")

        # Uncomment in moderngl 5.6+
        # mesh.instance(prog)

    def test_divisors(self):
        """Test defining buffers with different divisor types"""
        mesh = VAO("test", mode=moderngl.LINES)
        mesh.buffer(
            numpy.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0], dtype="f4"), "3f/v", "position"
        )
        mesh.buffer(
            numpy.array([0.0, 0.0, 1.0, 1.0, 0.0, 1.0], dtype="f4"), "3f/r", "normal"
        )
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0], dtype="f4"), "2f/i", "uv")

        buffer1 = mesh.get_buffer_by_name("position")
        buffer2 = mesh.get_buffer_by_name("normal")
        buffer3 = mesh.get_buffer_by_name("uv")

        attributes = ["position", "normal", "uv"]
        self.assertEqual(
            buffer1.content(attributes), (buffer1.buffer, "3f/v", "position")
        )
        self.assertEqual(
            buffer2.content(attributes), (buffer2.buffer, "3f/r", "normal")
        )
        self.assertEqual(buffer3.content(attributes), (buffer3.buffer, "2f/i", "uv"))



from unittest import TestCase

from moderngl_window.geometry import AttributeNames


class AttributeNamesTestCase(TestCase):

    def test_attributes(self):
        """Ensure we can create an instance with partial overrides"""
        attrs = AttributeNames(
            position="vertex",
            normal="normals",
        )
        self.assertEqual(attrs.POSITION, "vertex")
        self.assertEqual(attrs.NORMAL, "normals")
        self.assertEqual(attrs.TANGENT, "in_tangent")
